##
# File: TimeStamp.py
# Date: 4-Mar-2015
#
# Updates:
#   13-Apr-2015 - Added daylight time offset to correct for this.
#                 The manner of TZ incorporation does
#
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import time
import datetime
from dateutil import tz
from decimal import Decimal


class TimeStamp(object):
    def __init__(self, refDate="2000-01-01 00:00:00"):
        self.__useTz = True
        self.__refDate = refDate

        datetime_obj = datetime.datetime.strptime(self.__refDate, "%Y-%m-%d %H:%M:%S")

        if self.__useTz:
            utctz = tz.tzutc()
            self.__referenceDate = datetime_obj.replace(tzinfo=utctz)
        else:
            self.__referenceDate = datetime_obj

        self.__dstOffset = -self.__isDST() * 3600.0

    def __isDST(self):
        """Return 1 if the local TZ is on daylight time or 0 otherwise."""
        return time.localtime().tm_isdst

    def getTimeReference(self):
        """Return the datetime object for the time reference"""
        return self.__referenceDate

    def getSecondsFromReference(self):
        """Return the number of seconds since the time reference with micosecond precision."""
        if self.__useTz:
            utctz = tz.tzutc()
            tNow = datetime.datetime.utcnow().replace(tzinfo=utctz)
        else:
            tNow = datetime.datetime.utcnow()
        c = tNow - self.__referenceDate
        seconds = ((c.days * 24 * 60 * 60 + c.seconds) * 1000000 + c.microseconds) / 1000000.0
        return Decimal(repr(seconds))

    def getSecondsFromEpoc(self, secondsFromReference):
        """Return the number of seconds since the Epoc.

        Input: microseconds since the reference date (self.__referenceDate)
        """

        ref = time.mktime(self.__referenceDate.timetuple())
        d = float(ref) + float(secondsFromReference)
        return d

    def getTimeStringUTC(self, secondsFromReference):
        """Return the formatted time string in the UTC

        Input: seconds since the reference date (self.__referenceDate)
        """
        if secondsFromReference is None:
            return ""
        ref = time.mktime(self.__referenceDate.timetuple())

        fmt = "%Y-%m-%d %H:%M:%S %Z%z"

        d = float(ref) + float(secondsFromReference) + self.__dstOffset
        return datetime.datetime.fromtimestamp(int(d)).strftime(fmt)

    def getTimeStringLocal(self, secondsFromReference):
        """Return the formatted time string in the local time zone.

        Input: seconds since the reference date (self.__referenceDate)
        """
        if secondsFromReference is None:
            return ""
        ltz = tz.tzlocal()
        utctz = tz.tzutc()
        ref = time.mktime(self.__referenceDate.timetuple())

        fmt = "%Y-%m-%d %H:%M:%S %Z%z"

        d = float(ref) + float(secondsFromReference) + self.__dstOffset
        dt = datetime.datetime.fromtimestamp(int(d))
        dt = dt.replace(tzinfo=utctz)
        return dt.astimezone(ltz).strftime(fmt)

    def getTimeFromEpoc(self, secondsFromEpoc):
        """Return the formatted time string (UTC).

        Input: seconds since the epoc.
        """
        if secondsFromEpoc is None:
            return ""
        fmt = "%Y-%m-%d %H:%M:%S %Z%z"

        d = float(secondsFromEpoc)
        return datetime.datetime.fromtimestamp(int(d)).strftime(fmt)


if __name__ == "__main__":
    #
    tS = TimeStamp()
    print("getSecondsFromReference() %s" % tS.getSecondsFromReference())
    print("UTC (0)  %s" % tS.getTimeStringUTC(0.0))
    print("UTC (now)  %s" % tS.getTimeStringUTC(tS.getSecondsFromReference()))
    print("Local now  %s" % tS.getTimeStringLocal(tS.getSecondsFromReference()))
    #
    tS = TimeStamp()
    print("getSecondsFromReference() %s" % tS.getSecondsFromReference())
    print("getSecondsFromEpoc() %s" % tS.getSecondsFromEpoc(tS.getSecondsFromReference()))
    #
    print("tS.getTimeStringLocal(0) %s" % tS.getTimeStringLocal(0))
    print("tS.getTimeStringLocal(3600*24) %s" % tS.getTimeStringLocal(3600 * 24))
    print("")
    print("tS.getTimeStringUTC(0) %s" % tS.getTimeStringUTC(0))
    print("tS.getTimeStringUTC(3600*24) %s" % tS.getTimeStringUTC(3600 * 24))



##
# File:    ProcessManager.py
# Date:    15-Mar-2009
#
# Updates:
#  22-April-2010 : Incorporation into enterpise.
#  14-Feb-2015   jdw  -- initial refactoring -    gutted
#   9-Mar-2015   jdw remove WFEapplication dependency --
#  2-May-2015    jdw gutted -
#
##

"""
manages the application processes via the API
This is the class that manages processes - they
can be within the WFE (there are a small number)
or they can be within the API
"""

__docformat__ = "restructuredtext en"
__author__ = "Tom Oldfield"
__email__ = "oldfield@ebi.ac.uk"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.24"

import sys
import os
import time
from threading import Thread
import logging
import traceback


from wwpdb.apps.wf_engine.engine.ServerMonitor import ServerMonitor

from wwpdb.utils.wf.process.ProcessRunner import ProcessRunner
from wwpdb.apps.wf_engine.engine.EngineUtils import EngineUtils

# from wwpdb.apps.wf_engine.wf_engine_utils.run.MyLogger import MyLogger
from wwpdb.apps.wf_engine.wf_engine_utils.time.TimeStamp import TimeStamp

logger = logging.getLogger(name="root")


class ProcessManager(Thread):

    """
    The process manager runs a process and sits and waits

    """

    def __init__(self, task, debug=0):
        """Input task is the current task object"""
        Thread.__init__(self)
        self.debug = debug
        self.__verbose = debug
        self.status = "unknown"
        self.elapsedTime = None
        self.task = task
        self.result = -1
        self._thetime = 0
        # self.__myLog = MyLogger()
        self.depositionID = None
        self.WorkflowClassID = None
        self.WorkflowInstanceID = None
        #
        self.__timeStamp = TimeStamp()
        #
        # input data for process : these are key / value pairs - of the input data
        self.input = {}
        #
        self.__taskParameterD = {}
        #
        # output data for process : these are key / value pairs
        self.output = {}

        self.instanceExitState = "finished"

    def getExitState(self):
        return self.instanceExitState

    def setInput(self, valueD, taskParamD):
        if self.debug > 0:
            if valueD is not None:
                logger.info("+ProcessManager.setInput() value dictionary %r\n", valueD.items())
            if taskParamD is not None:
                logger.info("+ProcessManager.setInput() task parameter dictionary %r\n", taskParamD.items())
        self.__taskParameterD = taskParamD
        self.input = valueD

    def setOutput(self, valueD):
        if self.debug > 0:
            if valueD is not None:
                logger.info("+ProcessManager.setOutput() value dictionary %r\n", valueD.items())

        self.output = valueD

    def setStatsInfo(self, depositionID, WorkflowClassID, WorkflowInstanceID):

        self.depositionID = depositionID
        self.WorkflowClassID = WorkflowClassID
        self.WorkflowInstanceID = WorkflowInstanceID

    def abort(self):
        logger.error(" call to abort")
        self.status = "abort"

    def __setDBInstStatus(self, mode, eUtilObj):

        now = self.__timeStamp.getSecondsFromReference()
        instDB = {}
        instDB["WF_INST_ID"] = self.WorkflowInstanceID
        instDB["WF_CLASS_ID"] = self.WorkflowClassID
        instDB["DEP_SET_ID"] = self.depositionID
        instDB["STATUS_TIMESTAMP"] = now
        eUtilObj.updateStatus(instDB, mode)

        sql = (
            "update wf_instance_last set status_timestamp="
            + str(now)
            + ", inst_status='"
            + mode
            + "', wf_class_id = '"
            + self.WorkflowClassID
            + "', wf_inst_id = '"
            + self.WorkflowInstanceID
            + "' where dep_set_id = '"
            + self.depositionID
            + "'"
        )
        ok = eUtilObj.runUpdateSQL(sql)
        if ok < 1:
            logger.error("+ProcessManager.__setDBInstStatus() CRITICAL : failed to update wf_instance_last table")

    def __setDBTaskStatus(self, mode, eUtilObj):

        taskID = {}
        taskID["WF_TASK_ID"] = self.task.name
        taskID["WF_INST_ID"] = self.WorkflowInstanceID
        taskID["WF_CLASS_ID"] = self.WorkflowClassID
        taskID["DEP_SET_ID"] = self.depositionID
        taskID["STATUS_TIMESTAMP"] = self.__timeStamp.getSecondsFromReference()
        eUtilObj.updateStatus(taskID, mode)

    def run(self):
        """
        This is the entry point of the process manager which is
        legacy run as a Thread

        It then justs calls runProcess
        """
        eUtilObj = EngineUtils(verbose=self.__verbose)
        self.status = "starting"
        self.__setDBTaskStatus("running", eUtilObj)

        if self.debug > 1:
            logger.info("+ProcessManager.run : %s : %s", str(self.status), str(self.__getElapsedTime()))

        self.__runProcess(eUtilObj)

        if self.debug > 1:
            logger.info("+ProcessManager.run :  %s %s", str(self.status), str(self.__getElapsedTime()))

        self.elapsedTime = None
        return self.status

    def __waitTime(self, sec):
        time.sleep(sec)

    def getState(self):
        return self.status

    def getResult(self):
        return self.result

    def __updateElapsedTime(self):
        """Update the elapsed time."""
        self.elapsedTime = self.elapsedTime + (self.__timeStamp.getSecondsFromReference() - self._thetime)

    def __getElapsedTime(self):
        """Show the elapsed time of the thread."""
        if self.elapsedTime is None:
            self.elapsedTime = 0
            self._thetime = self.__timeStamp.getSecondsFromReference()
        else:
            self.__updateElapsedTime()
        return "%5.2fs" % self.elapsedTime

    def __runProcess(self, eUtilObj):
        """
        This is the action method called by "run" It does all the work.

        if the uniqueWhere name = wfe : then do the task here . mostly very simple
          about the only useful wfe function is "wait"

        if the uniqueWhere name = "api" : then do the task in the API using  ProcessRunner()
        """

        self.status = "running"
        #  All output is directed to this task logfile -

        processLogName = (
            str(self.depositionID)
            + "_"
            + str(self.WorkflowClassID)
            + "_"
            + str(self.task.name)
            + "_"
            + str(self.WorkflowInstanceID)
            + "_"
            + str(self.task.name)
            + "_"
            + str(self.task.uniqueName)
            + ".log"
        )
        processLog = os.path.join(eUtilObj.getLogDirectoryPath(self.depositionID), processLogName)
        logfh = open(processLog, "w")

        if self.debug > 0:
            logfh.write("+ProcessManager.__runProcess :  task name     = " + str(self.task.name) + "\n")
            logfh.write("+ProcessManager.__runProcess :  name (simple) = " + str(self.task.nameHumanReadable) + "\n")
            logfh.write("+ProcessManager.__runProcess :  uniqueAction  = " + str(self.task.uniqueAction) + "\n")
            logfh.write("+ProcessManager.__runProcess :  uniqueWhere   = " + str(self.task.uniqueWhere) + "\n")

        if self.task.uniqueWhere == "wfe":
            ##
            #  The following tasks link out to functions that should deprecated ---
            if self.task.uniqueAction == "initDB":
                self.instanceExitState = "init"
                eUtilObj.initDepositContext(self.depositionID, self.WorkflowClassID, self.WorkflowInstanceID, self.input)
            elif self.task.uniqueAction == "releaseHPUB":
                eUtilObj.setReleaseStatus(self.depositionID, "HPUB")
            elif self.task.uniqueAction == "sendDepEmail":
                eUtilObj.sendDepositorEmail(self.depositionID, self.task.description)
            elif self.task.uniqueAction == "randomAnnotator":
                eUtilObj.setRandomAnnotator(self.depositionID)
            #
            ##
            elif self.task.uniqueAction == "die" or self.task.uniqueAction == "throw":
                self.status = self.task.uniqueName
                return
            elif self.task.uniqueAction == "finished":
                eUtilObj.setFinished(self.depositionID)
            elif self.task.uniqueAction == "resetStatus":
                eUtilObj.resetInitialStateDB(self.depositionID)
            elif self.task.uniqueAction == "fileExist":
                eUtilObj.testFilesExist(self.input, self.output)
            elif self.task.uniqueAction == "wait":
                self.__waitTime(self.task.runTime)
            elif self.task.uniqueAction == "monitorDB":
                monitor = ServerMonitor(self.WorkflowInstanceID, self.WorkflowClassID, self.depositionID, self.debug)
                monitor.watch()
            else:
                logfh.write("+ProcessManager.__runProcess :   ***** Unknown WFE application ****\n")
                self.__setDBTaskStatus("badNameX", eUtilObj)
                self.__setDBInstStatus("exception", eUtilObj)
                sys.exit(1)
        elif self.task.uniqueWhere == "api":

            if self.debug > 0:
                process = ProcessRunner(True, logfh)
            else:
                process = ProcessRunner(False, logfh)
            process.setAction(self.task.uniqueAction)

            if self.__taskParameterD is not None:
                process.setParameterDict(self.__taskParameterD)

            if self.input is not None:
                for key, value in self.input.items():
                    if self.debug > 1:
                        try:
                            cN = value.__class__.__name__
                            logfh.write("+ProcessManager.__runProcess :  setting input %s using object type %s\n" % (key, cN))
                            value.printMe()
                        except Exception as _e:  # noqa: F841
                            logfh.write("+ProcessManager.__runProcess :  setting input %s with value %r\n" % (key, value))
                    process.setInput(key, value)
            if self.output is not None:
                for key, value in self.output.items():
                    if self.debug > 1:
                        try:
                            cN = value.__class__.__name__
                            logfh.write("+ProcessManager.__runProcess :  setting output %s using object type %s\n" % (key, cN))
                            value.printMe()
                        except Exception as _e:  # noqa: F841
                            logfh.write("+ProcessManager.__runProcess :  setting output %s with value %r\n" % (key, value))

                    process.setOutput(key, value)

            if not process.preCheck():
                logfh.write("+ProcessManager.__runProcess : data pre-check fails for task %s action %s" % (self.task.name, self.task.uniqueAction))
                self.status = "exception"
                self.__setDBTaskStatus("exception", eUtilObj)
                self.__setDBInstStatus("exception", eUtilObj)
                sys.exit(1)

            ok = process.run()
            if not ok:
                pass

        if self.debug > 0:
            for key, value in self.output.items():
                try:
                    cN = value.__class__.__name__
                    logfh.write("+ProcessManager.__runProcess :  returned output %s has object type %s\n" % (key, cN))
                    value.printMe(logfh)
                except Exception as _e:  # noqa: F841
                    traceback.print_exc(file=logfh)
                    logfh.write("+ProcessManager.__runProcess :  returned output %s has value %r\n" % (key, value))

        self.status = "complete"
        self.__setDBTaskStatus("complete", eUtilObj)

        if self.debug > 0:
            logfh.write("+ProcessManager.__runProcess :  finished task name  %s\n" % self.task.name)

        # close the log file
        logfh.close()



##
# File: WFEapplications.py
# Date:
#
# jdw    WARNING loads of domain specific functions here called from all over wfm and depui.
# jdw -- grab bag of functions -- change date 'unknown' date today
# jdw -- replace missing workflow with placeholder workflow in initilliseInstance(DBstatusAPI, id)
#
# 4-Aug-2015 jdw -- add def runOutOfBandWorkflow(DBstatusAPI, depSetId, classId, classFileName)
# Aug-2015 lm -- add def getWfStatus
#
#  14-Jul-2016 jdw add validation_server flag to functions runWorkflowOnUpload() & runValidationUpload()
#  26-Sep-2017 ep  initilliseDepositDict() parameterize strings to ensure proper quoting
#  26-Oct-2018 ep  remove dead code, switch to logging
#
# #
import time
import sys
import os
import pickle
import datetime
import logging

from wwpdb.utils.wf.process.ProcessRunner import ProcessRunner
from wwpdb.utils.wf.WfDataObject import WfDataObject
from wwpdb.utils.wf.dbapi.WfDbApi import WfDbApi
from wwpdb.utils.wf.dbapi.dbAPI import dbAPI
from wwpdb.utils.config.ConfigInfo import ConfigInfo, getSiteId
from wwpdb.utils.wf.dbapi.WFEtime import getTimeNow
from email.mime.text import MIMEText

logger = logging.getLogger()


def getPicklePath(depSetId=None):
    """Return path to deposition pkl files - to break circular dependency
    if depSetId set, return deposition specific path
    """

    # logger.debug("Getting PicklePath for %s" % depSetId)

    cI = ConfigInfo()
    WWPDB_APP_VERSION_STRING = "v-200"

    depstrpath = cI.get("SITE_ARCHIVE_STORAGE_PATH")
    if not depstrpath:
        logger.error("SITE_ARCHIVE_STORAGE_PATH not set")
        return None

    file_upload_temp_dir = os.path.join(depstrpath, "deposit", "temp_files")
    storage_pickled_depositions = os.path.join(file_upload_temp_dir, "deposition" + "-" + WWPDB_APP_VERSION_STRING)

    if depSetId:
        fPath = os.path.join(storage_pickled_depositions, str(depSetId))
    else:
        fPath = storage_pickled_depositions

    return fPath


def getNotesEmailAddr():
    """Returns the email address for archiving messages"""
    email = "wwpdb-da-system@rcsb.rutgers.edu"

    siteid = getSiteId()
    # Debugging at PDBe
    if siteid == "PDBE_DEV":
        cI = ConfigInfo()
        email = cI.get("ANNOTATION_EMAIL")

    return email


def runOutOfBandWorkflow(DBstatusAPI, depSetId, classId, classFileName):
    """UPDATE status.communication SET  sender=%s, receiver=%s, wf_class_id=%s, wf_class_file=%s,
                                  command=%s, status=%s, actual_timestamp=%s, data_version=%s
                                  WHERE ( dep_set_id=%s);

    ['WFUTILS', 'WFE', 'PDBX2PDBX_DEP', 'wf_op_pdbx2pdbx_fs_deposit.xml',
     'runWF', 'PENDING', Decimal('492013055.527617'), 'latest',
     'D_1000208939']
    """
    t = getTimeNow()
    sqlTemplate = "UPDATE status.communication SET  sender='%s', receiver='%s', wf_class_id='%s', wf_class_file='%s', command='%s', status='%s', actual_timestamp='%s', data_version='%s' WHERE ( dep_set_id='%s')"  # noqa: E501
    tdata = ("WFUTILS", "WFE", classId, classFileName, "RunWF", "PENDING", str(t), "latest", depSetId)
    sql = sqlTemplate % (tdata)
    ok = DBstatusAPI.runUpdateSQL(sql)
    return ok


def reRunWorkflow(depID):
    """
    method to reRun the validation WF just by setting the communication status to pending
    """

    status = "OK"
    try:
        timestamp = getTimeNow()
        wfApi = WfDbApi(verbose=True)
        sql = "select wf_class_file,command from communication where dep_set_id = '" + str(depID) + "'"
        rows = wfApi.runSelectSQL(sql)
        for row in rows:
            if row[0] in ["ValidDeposit.xml", "depRunOnUpload.xml", "wf_op_validdeposit_fs_deposit.xml", "wf_op_uploaddep_fs_deposit.xml"] and row[1] in ["runWF", "restartGoWF"]:
                sql = "update communication set status = 'PENDING',actual_timestamp = '" + str(timestamp) + "' where dep_set_id = '" + str(depID) + "'"
                ok = wfApi.runUpdateSQL(sql)
                if ok < 1:
                    status = "Failed to mark this for restart "
            else:
                status = "WF was not in deposition "
            break
    except Exception as e:
        status = "Kill WF error " + str(e)

    logger.info("Restart WF request:  %s : %s", str(depID), str(status))

    return status


def getValidationStatus(depID, wfApi=None):
    """
    method to see if there is a valid row in the database for communication and that it is WORKING for deposition uploads
    """

    if not wfApi:
        wfApi = WfDbApi(verbose=True)

    sql = (
        "select status from communication where dep_set_id = '"
        + str(depID)
        + "' and wf_class_file in ('ValidDeposit.xml','depRunOnUpload.xml','wf_op_validdeposit_fs_deposit.xml','wf_op_uploaddep_fs_deposit.xml')"
    )
    rows = wfApi.runSelectSQL(sql)
    logger.debug(sql)
    logger.debug(str(rows))

    if rows is not None and len(rows) > 0:
        for row in rows:
            return row[0].lower()
    else:
        return "unknown"

    return "unknown"


def getWfStatus(depID, wfApi=None):
    """
    method to check the status of a current workflow; and to see if there is a valid row in the database for communication and that it is WORKING for deposition uploads
    """

    if not wfApi:
        wfApi = WfDbApi(verbose=True)

    # sql = "select status from communication where dep_set_id = '" + str(depID) + "' and wf_class_file = '" + classFile +"'"
    sql = "select status, wf_class_file from communication where dep_set_id = '" + str(depID) + "'"
    rows = wfApi.runSelectSQL(sql)
    if rows is not None and len(rows) > 0:
        # if rows[0][0] != 'WORKING':
        if rows[0][0] == "FINISHED":
            logger.info("wf_engine.engine.WFEapplications.getWfStatus: %s", sql)
            logger.info("wf_engine.engine.WFEapplications.getWfStatus: %s", str(rows))
        for row in rows:
            return row
            # return row[0].lower()
    else:
        return "unknown"

    return "unknown"


def killAllWF(depID, who):

    status = "OK"
    try:
        timestamp = getTimeNow()
        wfApi = WfDbApi(verbose=True)

        sql = (
            "update communication set command = 'killWF', actual_timestamp = '" + str(timestamp) + "', receiver = 'WFE', status = 'PENDING' where dep_set_id = '" + str(depID) + "'"
        )
        ok = wfApi.runUpdateSQL(sql)
        if ok < 1:
            status = "Failed to mark this for process kill"
    except Exception as e:
        status = "Kill WF error " + str(e)

    logger.info("Kill WF request:  %s from %s", str(depID), str(who))

    return status


def waitTime(sec):

    time.sleep(sec)


# Updated jdw
#


def wfClassDir():
    # JDW -- deprecated -- directory containing the directory containing ---
    siteId = getSiteId(defaultSiteId="WWPDB_DEPLOY_TEST")
    cI = ConfigInfo(siteId)
    ret = cI.get("SITE_WF_PYTHON_PATH")
    # ret = os.path.abspath(os.path.join(os.path.abspath(__file__), '../../')) + '/'
    return ret


def getNextInstanceId(DBstatusAPI, depID, prt=sys.stderr):
    """
    Create new workflow instance from the statusDB
    Returns the ID
    The status API returns an ID as a number, result is padded with zeros on a field width of 3
    """
    sql = "select max(cast(substr(wf_inst_id,3) as decimal(5,0)))+1 from wf_instance where dep_set_id = '" + str(depID) + "'"

    ll = DBstatusAPI.runSelectSQL(sql)

    if ll is None:
        instID = 1
    else:
        for el in ll:
            if el is None:
                instID = 1
            else:
                if el[0] is None:
                    instID = 1
                else:
                    instID = el[0]

    ret = "W_" + str(instID).zfill(3)

    prt.write("+getNextInstanceId: Created new workflow instance ID = %s" % str(ret))

    return ret


def insertInitialStateDB(DBstatusAPI, depID, debug=0, prt=sys.stderr):
    """Method to insert or update the instance of the annotation workflow within the wf_instance/wf_instance_last table -"""
    # Check if this workflow already exists in the wf_instance table -
    #
    sql = "select ordinal from wf_instance where (wf_class_id = 'Annotate')  and (dep_set_id = '" + str(depID) + "')  order by status_timestamp desc limit 1"
    allList = DBstatusAPI.runSelectSQL(sql)

    if allList is not None and len(allList) > 0:
        return updateInitialStateDB(DBstatusAPI, depID, debug=debug, prt=prt)

    now = getTimeNow()
    instId = getNextInstanceId(DBstatusAPI, depID, prt=prt)

    sql = (
        "insert into wf_instance (wf_inst_id, wf_class_id, dep_set_id, owner, inst_status, status_timestamp) values ('"
        + instId
        + "','Annotate','"
        + depID
        + "','Annotation.bf.xml','init','"
        + str(now)
        + "')"
    )
    ok = DBstatusAPI.runInsertSQL(sql)
    if int(ok) == 1:
        prt.write("+insertInitialStateDB - insert returns %r row into wf_instance for %r\n" % (ok, depID))
    #
    # Try update then insert -
    sql = (
        "update wf_instance_last set owner = 'Annotation.bf.xml', wf_class_id = 'Annotate' , inst_status = 'init', "
        + " wf_inst_id = '"
        + instId
        + "', status_timestamp = '"
        + str(now)
        + "'  where dep_set_id = '"
        + str(depID)
        + "'"
    )
    prt.write("+insertInitialStateDB - SQL =  %s \n" % sql)
    ok = DBstatusAPI.runUpdateSQL(sql)
    prt.write("+insertInitialStateDB - return ok  =  %r \n" % ok)
    if int(ok) == 1:
        prt.write("+updateInitialStateDB - Updating Annotate class in wf_instance suceeded for %s\n" % depID)
    else:
        sql = (
            "insert into wf_instance_last (wf_inst_id, wf_class_id, dep_set_id, owner, inst_status, status_timestamp) values ('"
            + instId
            + "','Annotate','"
            + depID
            + "','Annotation.bf.xml','init','"
            + str(now)
            + "')"
        )
        ok = DBstatusAPI.runInsertSQL(sql)
        if int(ok) == 1:
            prt.write("+insertInitialStateDB - Inserted 1 row into wf_instance_last for %s\n " % depID)
        else:
            prt.write("+insertInitialStateDB - Updating Annotate class in wf_instance failed for %s\n" % depID)


def resetInitialStateDB(DBstatusAPI, depID, debug=0, prt=sys.stderr):
    """
    Method to reset (insert/update) a annotation workflow back to init state by setting the owner, status and class
    """
    #
    sql = "select ordinal from wf_instance where (wf_class_id = 'Annotate')  and (dep_set_id = '" + str(depID) + "')  order by status_timestamp desc limit 1"
    allList = DBstatusAPI.runSelectSQL(sql)

    if (allList is not None) and (len(allList) > 0):
        return updateInitialStateDB(DBstatusAPI, depID, debug=debug, prt=prt)
    else:
        return insertInitialStateDB(DBstatusAPI, depID, debug=debug, prt=prt)


def updateInitialStateDB(DBstatusAPI, depID, debug=0, prt=sys.stderr):  # pylint: disable=unused-argument
    """
    Method to reset the annoation workflow back to init state by setting the owner, status and class
    """
    #
    sql = "select ordinal from wf_instance where (wf_class_id = 'Annotate')  and (dep_set_id = '" + str(depID) + "')  order by status_timestamp desc limit 1"
    allList = DBstatusAPI.runSelectSQL(sql)

    if allList is not None:
        for allRow in allList:
            now = getTimeNow()
            ordinal = allRow[0]
            prt.write("+updateInitialStateDB - Ordinal  %s found for depID %s \n" % (depID, ordinal))
            sql = (
                "update wf_instance set owner = 'Annotation.bf.xml', wf_class_id = 'Annotate' , inst_status = 'init', status_timestamp = '"
                + str(now)
                + "' where ordinal = "
                + str(ordinal)
            )
            ok = DBstatusAPI.runUpdateSQL(sql)
            # update the wf_instance_last table
            sql = (
                "update wf_instance_last set owner = 'Annotation.bf.xml', wf_class_id = 'Annotate' , inst_status = 'init', status_timestamp = '"
                + str(now)
                + "' where dep_set_id = '"
                + str(depID)
                + "'"
            )
            ok = DBstatusAPI.runUpdateSQL(sql)
            if int(ok) == 1:
                prt.write("+updateInitialStateDB - %s status=%r for updating Annotate class in wf_instance \n" % (depID, ok))
            else:
                prt.write("+updateInitialStateDB - %s status=%r for updating Annotate class in wf_instance \n" % (depID, ok))


def DepUIgetDepositorEmail(depID):
    """
    go get the depositor email
    """

    logger.info("Getting email address for '%s'", depID)
    # This code is here to break a dependence on DepUI code from WFE
    fPath = getPicklePath(depID)

    if not fPath:
        logger.error(" The storage location of the pickle file is not known")
        return None

    fName = os.path.join(fPath, "formdata.pkl")

    if os.path.isfile(fName):
        try:
            f = open(fName, "rb")
            dat = pickle.load(f)
            f.close()
            if "email" in dat:
                return dat["email"]
            else:
                return None
        except Exception as e:
            logger.exception("Tried to get the formdata and return email %s", str(e))
            return None
    else:
        ss = dbAPI(depID)
        ret = ss.runSelectNQ(table="user_data", select=["email", "role"], where={"dep_set_id": depID, "role": "valid"})
        for r in ret:
            return r[0]
        return None


def WFEsendEmail(email, frm, subject, message, bcc=None):
    # Used when enabling FTP from WFM
    import smtplib

    if email is None:
        logger.error(" Invalid email %s", str(email))
        return

    msg = MIMEText(message)
    msg["Subject"] = subject
    msg["From"] = frm
    msg["To"] = email
    sendTo = [email]
    if bcc:
        sendTo = [email] + [bcc]
        logger.debug("send email (bcc) to %s", str(bcc))

    cI = ConfigInfo()
    noreplyaddr = cI.get("SITE_NOREPLY_EMAIL", "noreply@mail.wwpdb.org")

    # Send the message via our own SMTP server, but don't include the
    # envelope header.
    logger.debug("send email from %s", str(frm))
    logger.debug("send email to %s", str(email))
    try:
        s = smtplib.SMTP("localhost")
        s.sendmail(noreplyaddr, sendTo, msg.as_string())
        s.quit()
    except Exception as e:
        logger.error(" Exception in WFE.WFEsendEmail %s", str(e))


def resetComms(DBstatusAPI, id):  # pylint: disable=redefined-builtin

    t = getTimeNow()
    sql = (
        "update communication set sender='INSERT',receiver='LOAD',command='INIT',status='INIT',actual_timestamp='"
        + str(t)
        + "',parent_dep_set_id='"
        + str(id)
        + "',parent_wf_class_id='Annotate',wf_class_file = 'Annotation.bf.xml',parent_wf_inst_id='W_001' where dep_set_id = '"
        + id
        + "'"
    )
    logger.info(sql)
    DBstatusAPI.runUpdateSQL(sql)


def initialiseComms(DBstatusAPI, id):  # pylint: disable=redefined-builtin

    sql = "select dep_set_id from communication where dep_set_id = '" + str(id) + "'"

    allList = DBstatusAPI.runSelectSQL(sql)

    logger.info("****************%s", str(allList))

    if allList is None or len(allList) < 1:
        # The normal case - insert a new value
        t = getTimeNow()
        sql = (
            "insert communication (sender,receiver,dep_set_id,command,status,actual_timestamp,parent_dep_set_id,parent_wf_class_id,parent_wf_inst_id,wf_class_file) values ('INSERT','LOAD','"  # noqa: E501
            + str(id)
            + "','INIT','INIT',"
            + str(t)
            + ", '"
            + str(id)
            + "','Annotate','W_001','Annotation.bf.xml')"
        )
        logger.info(sql)
        DBstatusAPI.runInsertSQL(sql)
    else:
        # we already have that dep_setID - so just reset the communication
        resetComms(DBstatusAPI, id)


def initilliseInstance(DBstatusAPI, id):  # pylint: disable=redefined-builtin
    """If there is no existing wf_instance for the input id then add a placeholder instance -"""
    sql = "select dep_set_id from wf_instance where dep_set_id = '" + str(id) + "'"

    allList = DBstatusAPI.runSelectSQL(sql)

    t = getTimeNow()
    if allList is None or len(allList) < 1:
        # The normal case - insert a new value
        sql = (
            "insert wf_instance (wf_inst_id,wf_class_id,dep_set_id,owner,inst_status,status_timestamp) values ('W_001','PLACEHOLDER','"
            + str(id)
            + "','wf_op_placeholder.xml','dep','"
            + str(t)
            + "')"
        )
        DBstatusAPI.runInsertSQL(sql)
        sql = (
            "insert wf_instance_last (wf_inst_id,wf_class_id,dep_set_id,owner,inst_status,status_timestamp) values ('W_001','PLACEHOLDER','"
            + str(id)
            + "','wf_op_placeholder.xml','dep','"
            + str(t)
            + "')"
        )
        DBstatusAPI.runInsertSQL(sql)
    else:
        # we already have that dep_setID - so just reset the communication
        pass


def initilliseDepositDict(DBstatusAPI, id, pdb="?", data=None):  # pylint: disable=unused-argument,redefined-builtin
    """
    raw deposition table inialisation
    """

    depDB = {}
    depDB["DEP_SET_ID"] = id

    if DBstatusAPI.exist(depDB):
        logger.info("ID exists ")
        sql = "update deposition set "
        args = ()
        first = False
        for key, value in data.items():
            if not first:
                first = True
            else:
                sql = sql + ","
            sql = sql + str(key) + " = %s"
            # comma important with single value tuples
            args += (str(value),)
        sql = sql + " where dep_set_id = %s"
        args += (str(id),)

        logger.info("%s %s", sql, args)
        DBstatusAPI.runUpdateSQL(sql, args)
    else:
        # I do not think this code path is used - missing
        # commas between keys and values - EP 2017-09-26
        logger.info("ID is new")
        sql = "insert deposition (dep_set_id,"
        for key, value in data.items():
            sql = sql + str(key)
        sql = sql + ') values ("' + str(id) + '",'
        for key, value in data.items():
            sql = sql + '"' + str(value) + '"'

        sql = sql + ")"
        logger.info(sql)
        DBstatusAPI.runInsertSQL(sql)


def getdepUIPassword(DBstatusAPI, id):  # pylint: disable=redefined-builtin

    depDB = {}
    depDB["DEP_SET_ID"] = id

    try:
        if DBstatusAPI.exist(depDB):
            logger.info("ID exists ")
            sql = "select depPW from  deposition where dep_set_id = '" + str(id) + "'"
            logger.info(sql)
            allList = DBstatusAPI.runSelectSQL(sql)

            if allList is not None:
                for allRow in allList:
                    pw = allRow[0]
                    return pw
    except Exception as _e:  # noqa: F841
        logger.exception("Failed to get depUI PW in status DB")

    return None


def setdepUIPassword(DBstatusAPI, id, pw):  # pylint: disable=redefined-builtin
    """
    keep this separate so we never overwrite the PW
    """

    depDB = {}
    depDB["DEP_SET_ID"] = id

    try:
        if DBstatusAPI.exist(depDB):
            sql = "update deposition set depPW = '" + str(pw) + "' where dep_set_id = '" + str(id) + "'"
            DBstatusAPI.runUpdateSQL(sql)
        else:
            logger.info(" Cannot set depUI password - depID does not exist")
    except Exception as e:
        logger.exception("Failed to set depUI PW in status DB %s", str(e))


# API in use - references id by name...
def initilliseDepositV2(
    DBstatusAPI,
    id,  # pylint: disable=redefined-builtin,unused-argument
    pdb="?",
    date=None,  # pylint: disable=unused-argument
    initials="unknown",  # pylint: disable=unused-argument
    deposit_site="?",
    process_site="?",
    status_code="DEP",
    author_code="?",
    title="?",
    author_list="?",
    expt="?",
    status_code_exp="?",
    SG_center="?",
    ann="dep",
    email="",
    requested_codes=None,
):
    """
    raw deposition table inialisation
    """

    if requested_codes is None:
        requested_codes = []

    depDB = {}
    depDB["DEP_SET_ID"] = id

    if DBstatusAPI.exist(depDB):
        logger.info("ID exists ")
        sql = "update deposition set process_site = '" + str(process_site) + "', exp_method = '" + str(expt) + "' where dep_set_id = '" + str(id) + "'"
        logger.info(sql)
        DBstatusAPI.runUpdateSQL(sql)
    else:
        sql = (
            "insert deposition (dep_set_id, pdb_id, emdb_id, bmrb_id, deposit_site, process_site, status_code, author_release_status_code, title, author_list, exp_method, status_code_exp, SG_center, annotator_initials,email) values ('"  # noqa: E501
            + str(id)
            + "','"
            + str(pdb)
            + "','"
            + str(pdb)
            + "','"
            + str(pdb)
            + "','"
            + str(deposit_site)
            + "','"
            + str(process_site)
            + "','"
            + str(status_code)
            + "','"
            + str(author_code)
            + "','"
            + str(title)
            + "','"
            + str(author_list)
            + "','"
            + str(expt)
            + "','"
            + str(status_code_exp)
            + "','"
            + str(SG_center)
            + "','"
            + str(ann)
            + "','"
            + str(email)
            + "')"
        )

        DBstatusAPI.runInsertSQL(sql)
        if "EMDB" in requested_codes:
            sql = "update deposition set status_code_emdb = %r where dep_set_id = %r" % (status_code, id)
            DBstatusAPI.runInsertSQL(sql)


def runValidationUpload(DBstatusAPI, id, validation_server=False):  # pylint: disable=redefined-builtin
    """
    (old: Add request to run WF  : DepValModule.xml)
    Add request to run WF  : wf_op_validdeposit_fs_deposit.xml

    Jul/2016 function is called by deposit/depui/upload.py
    """

    status = getValidationStatus(id, wfApi=DBstatusAPI)
    now = getTimeNow()

    logger.info(" Engine.WFEapplications.runValidationUpload %s", str(now))
    logger.info(" Engine.WFEapplications.runValidationUpload %s", str(id))
    logger.info(" Engine.WFEapplications.runValidationUpload %s", str(status))

    if validation_server:
        wfName = "wf_op_validserver_fs_deposit.xml"
    else:
        wfName = "wf_op_validdeposit_fs_deposit.xml"

    if status == "working":
        sql = (
            "update communication set command = 'restartGoWF',  wf_class_file = '"
            + wfName
            + "', wf_class_id = 'DepVal', status = 'PENDING', actual_timestamp = '"
            + str(now)
            + "' where dep_set_id = '"
            + id
            + "'"
        )
    else:
        sql = (
            "update communication set sender = 'DEP', receiver = 'WFE', wf_class_file = '"
            + wfName
            + "', wf_class_id = 'DepVal', command = 'runWF', status = 'PENDING', actual_timestamp = '"
            + str(now)
            + "', parent_dep_set_id = '"
            + str(id)
            + "', parent_wf_class_id = 'DepUpload', parent_wf_inst_id = 'W_001' where dep_set_id = '"
            + id
            + "'"
        )

    logger.info(" Engine.WFEapplications.runValidationUpload - %s", str(sql))

    nrow = DBstatusAPI.runUpdateSQL(sql)
    logger.info(" WFE.runValidationUpload :  Rows affected: %i", nrow)


def runWorkflowOnUpload(DBstatusAPI, id, validation_server=False):  # pylint: disable=redefined-builtin
    """
    (old: Add request to run WF  : depRunOnUpload.xml)
    Add request to run WF  : wf_op_uploaddep_fs_deposit.xml

    Jul/2016 function is called by deposit/depui/upload.py
    """

    try:
        status = getValidationStatus(id, wfApi=DBstatusAPI)
        now = getTimeNow()
        logger.info(" Engine.WFEapplications.runWorkflowOnUpload %s", str(now))
        logger.info(" Engine.WFEapplications.runWorkflowOnUpload %s", str(id))
        logger.info(" Engine.WFEapplications.runWorkflowOnUpload %s", str(status))
        if validation_server:
            wfName = "wf_op_uploadvalsrv_fs_deposit.xml"
        else:
            wfName = "wf_op_uploaddep_fs_deposit.xml"
        if status == "working":
            # sql = "update communication set command = 'restartGoWF',  wf_class_file = 'depRunOnUpload.xml', status = 'PENDING', actual_timestamp = '" + \
            #    str(now) + "' where dep_set_id = '" + id + "'"
            sql = (
                "update communication set command = 'restartGoWF',  wf_class_file = '"
                + wfName
                + "', wf_class_id = 'uploadMod', status = 'PENDING', actual_timestamp = '"
                + str(now)
                + "' where dep_set_id = '"
                + id
                + "'"
            )

        else:

            # sql = "update communication set sender = 'DEP', receiver = 'WFE', wf_class_file = 'depRunOnUpload.xml', command = 'runWF', status = 'PENDING', actual_timestamp = '" + \
            #    str(now) + "', parent_dep_set_id = '" + str(id) + "', parent_wf_class_id = 'DepUpload', parent_wf_inst_id = 'W_001' where dep_set_id = '" + id + "'"
            sql = (
                "update communication set sender = 'DEP', receiver = 'WFE', wf_class_file = '"
                + wfName
                + "', wf_class_id = 'uploadMod', command = 'runWF', status = 'PENDING', actual_timestamp = '"
                + str(now)
                + "', parent_dep_set_id = '"
                + str(id)
                + "', parent_wf_class_id = 'DepUpload', parent_wf_inst_id = 'W_001' where dep_set_id = '"
                + id
                + "'"
            )

        logger.info(" Engine.WFEapplications.runWorkflowOnUpload - %s", str(sql))

        nrow = DBstatusAPI.runUpdateSQL(sql)
        logger.info(" WFE.runWorkflowOnUpload :  Rows affected: %i", nrow)
    except Exception as e:
        logger.exception("Exception Engine.WFEapplications.runWorkflowOnUpload %s", str(e))


def WFEsetAnnotator(DBstatusAPI, depID, annotator):

    sql = "update deposition set annotator_initials = '" + annotator + "' where dep_set_id = '" + depID + "'"
    _ok = DBstatusAPI.runUpdateSQL(sql)  # noqa: F841


def WFEexception(DBstatusAPI, depID):

    WFEsetCommunication(DBstatusAPI, "EXCEPTION", "EXCEPTION", depID)


def WFEfinished(DBstatusAPI, depID):

    WFEsetCommunication(DBstatusAPI, "FINISHED", "FINISHED", depID)


def WFEsetCommunication(DBstatusAPI, status, activity, depID):

    timeNow = getTimeNow()
    sql = (
        "update communication set status = '"
        + str(status)
        + "', activity = '"
        + str(activity)
        + "', actual_timestamp = "
        + str(timeNow)
        + " where dep_set_id  = '"
        + str(depID)
        + "'"
    )

    ok = DBstatusAPI.runUpdateSQL(sql)

    if ok == 1:
        logger.info("Set the communication to %s doe dep_set_id = %s", str(status), str(depID))
    else:
        logger.info("Failed to set the communication table to %s for dep_set_id = %s", str(status), str(depID))


def getInitialDate(depid):

    ret = "error"
    if depid is not None:
        wfApi = WfDbApi(verbose=True)
        sql = "select initial_deposition_date from deposition where dep_set_id = '" + depid + "'"
        dat = wfApi.runSelectSQL(sql)
        if dat is not None and len(dat) > 0:
            for r in dat:
                ret = r[0]

    return ret


def getAnnotatorInitials(depid):

    ret = "error"
    if depid is not None:
        wfApi = WfDbApi(verbose=True)
        sql = "select annotator_initials from deposition where dep_set_id = '" + depid + "'"
        dat = wfApi.runSelectSQL(sql)
        if dat is not None and len(dat) > 0:
            for r in dat:
                ret = r[0]

    return ret


def getStatusCode(depid):

    ret = "error"
    if depid is not None:
        wfApi = WfDbApi(verbose=True)
        sql = "select status_code from deposition where dep_set_id = '" + depid + "'"
        dat = wfApi.runSelectSQL(sql)
        if dat is not None and len(dat) > 0:
            for r in dat:
                ret = r[0]

    return ret


def getStatusCodeEMDB(depid):

    ret = "error"
    if depid is not None:
        wfApi = WfDbApi(verbose=True)
        sql = "select status_code_emdb from deposition where dep_set_id = '" + depid + "'"
        dat = wfApi.runSelectSQL(sql)
        if dat is not None and len(dat) > 0:
            for r in dat:
                ret = r[0]

    return ret


def setStatusCode(depID, code):

    valid = ["PROC", "AUTH", "DEP", "VAL", "HPUB", "REL", "REPL", "HOLD", "OBS", "WAIT", "WDRN", "REFI"]

    if code not in valid:
        return False, "Invalid status code"

    try:
        wfApi = WfDbApi(verbose=True)
        sql = "update deposition set status_code = '" + str(code) + "' where dep_set_id = '" + depID + "'"
        n = wfApi.runUpdateSQL(sql)
        if n == 0:
            return False, "Update did not occur - possible invalid dep_set_id or no change to code "
        else:
            return True, "Status updated"
    except Exception as e:
        return False, "Exception : " + str(e)


# ############################################
#  Believed not in use - except for examples
# ############################################

# Only used by examples - not in production
def populateDeposit(DBstatusAPI, depositionID):
    """
    Entry point added to the workflow system to create an object within
    the state system
    """

    debug = 3
    log = sys.stderr

    # Get the data out of the mmCIF
    if debug > 0:
        process = ProcessRunner(True, log)
    else:
        process = ProcessRunner(False, log)

    process.setAction("status")

    wfoInp = WfDataObject()
    wfoInp.setDepositionDataSetId(depositionID)
    wfoInp.setStorageType("archive")
    wfoInp.setContentTypeAndFormat("model", "pdbx")
    wfoInp.setVersionId("latest")
    # dP = wfoInp.getDirPathReference()
    # fP = wfoInp.getFilePathReference()
    # oVn = wfoInp.getFileVersionNumber()
    process.setInput("src", wfoInp)

    wfoOut = WfDataObject()
    wfoOut.setContainerTypeName("dict")
    wfoOut.setValueTypeName("string")
    process.setOutput("dst", wfoOut)

    if not process.preCheck():
        logger.error("WFE.WFEapplications.initialliseDeposit  : failed to setup process")

    if not process.run():
        logger.error("WFE.WFEapplications.initialliseDeposit : failed to run process")

    dataDict = wfoOut.getValue()

    if dataDict is not None:
        if "title" in dataDict:
            title = dataDict["title"].replace("'", " ")
        else:
            title = "Unknown"

        if "audit_author" in dataDict:
            auth = dataDict["audit_author"]
            if isinstance(auth, list):
                author_list = ""
                author_list = " ".join(["%s," % a for a in auth])
            elif isinstance(auth, str):
                author_list = auth
            else:
                logger.info(" Unknown type of author list %s", str(type(auth)))
            author_list = author_list.replace("'", " ")
        else:
            author_list = "unknown"

        if "exp_method" in dataDict:
            ex = dataDict["exp_method"]
            expt = ""
            if "status_code_nmr" in dataDict or ex.find("NMR") >= 0:
                expt += "NMR"
            if "status_code_em" in dataDict or ex.find("EM") >= 0:
                if len(expt) > 0:
                    expt += ","
                expt += "EM"
            if "status_code_sf" in dataDict or ex.find("XRAY") >= 0:
                if len(expt) > 0:
                    expt += ","
                expt += "xray"
        else:
            expt = "Unknown"

        if "full_name_of_center" in dataDict:
            sg_center = dataDict["full_name_of_center"]
        else:
            sg_center = "Unknown"

        now = datetime.datetime.utcnow().strftime("%Y-%m-%d")

        sql = (
            "update deposition set initial_deposition_date = '"
            + str(now)
            + "',title = '"
            + str(title)
            + "', author_list = '"
            + str(author_list)
            + "', exp_method = '"
            + str(expt)
            + "',sg_center = '"
            + str(sg_center)
            + "' where dep_set_id = '"
            + str(depositionID)
            + "' and status_code = 'DEP'"
        )
        logger.info(sql)
        nrow = DBstatusAPI.runUpdateSQL(sql)
        logger.info("rows change %s", str(nrow))


# NOT IN USE EXCEPT IN A SAMPLE SCRIPT in WFE XXXXX
def initilliseDB(metaData, depositionID, WorkflowClassID, WorkflowInstanceID, DBstatusAPI, dinput, debug=0, prt=sys.stderr):
    """
    method to extract information from the domain data
    using a depositionDB and put this into the statusDB
    """

    prt.write("WFE.WFEapplications.initialliseDB : ============================================\n")
    prt.write("WFE.WFEapplications.initialliseDB :  Intiallising DB for depID = " + str(depositionID) + "\n")
    prt.write("WFE.WFEapplications.initialliseDB : ============================================\n")

    if prt == sys.stderr:
        log = sys.stderr
    else:
        processLog = "log/status_" + str(depositionID) + "_" + WorkflowClassID + "_" + WorkflowInstanceID + ".log"
        log = open(processLog, "w")

    if debug > 0:
        process = ProcessRunner(True, log)
    else:
        process = ProcessRunner(False, log)

    process.setAction("status")

    if dinput is None:
        prt.write("WFE.WFEapplications.initialliseDB :  Requires input object \n")
        sys.exit(0)

    for key, value in dinput.items():
        if debug > 0:
            prt.write("WFE.WFEapplications.initialliseDB :  setting input " + str(key) + str(value.getReferenceType()) + "\n")
            prt.write("WFE.WFEapplications.initialliseDB :  setting input " + str(key) + str(value) + "\n")
        process.setInput("src", value)

    wfoOut = WfDataObject()
    wfoOut.setContainerTypeName("dict")
    wfoOut.setValueTypeName("string")
    process.setOutput("dst", wfoOut)

    if not process.preCheck():
        prt.write("WFE.WFEapplications.initialliseDB  : failed to setup process\n")

    if not process.run():
        prt.write("WFE.WFEapplications.initialliseDB : failed to run process\n")

    dataDict = wfoOut.getValue()
    if dataDict is None:
        logger.error(" There is no dataDictionary ")
        exit(0)

    depDB = {}
    depDB["DEP_SET_ID"] = depositionID
    depDB["PDB_ID"] = "????"
    if "accessions" in dataDict:
        acc = dataDict["accessions"]
        depDB["PDB_ID"] = "????"
        if acc is not None:
            accPDB = acc["PDB"]
            if accPDB is not None:
                if len(accPDB) > 0:
                    depDB["PDB_ID"] = accPDB[0]

    if "recvd_initial_deposition_date" in dataDict:
        depDB["INITIAL_DEPOSITION_DATE"] = dataDict["recvd_initial_deposition_date"]
    else:
        # JDW JDW
        today = datetime.datetime.utcnow().strftime("%Y-%m-%d")
        # depDB['INITIAL_DEPOSITION_DATE'] = 'unknown'
        depDB["INITIAL_DEPOSITION_DATE"] = today
    if "deposit_site" in dataDict:
        depDB["DEPOSIT_SITE"] = dataDict["deposit_site"]
    else:
        depDB["DEPOSIT_SITE"] = "unknown"
    if "process_site" in dataDict:
        depDB["PROCESS_SITE"] = dataDict["process_site"]
    else:
        depDB["PROCESS_SITE"] = "unknown"
    if "status_code" in dataDict:
        depDB["STATUS_CODE"] = dataDict["status_code"]
    else:
        depDB["STATUS_CODE"] = "PROC"
    logger.info("******* OVERWRITE OF STATUS CODE to PROC ***************")
    depDB["STATUS_CODE"] = "PROC"
    if "annotator_initials" not in dataDict or dataDict["annotator_initials"] == "UNASSIGNED":
        depDB["ANNOTATOR_INITIALS"] = "unknown"
    else:
        depDB["ANNOTATOR_INITIALS"] = dataDict["annotator_initials"]
        logger.info("******* OVERWRITE OF ANNOTATOR INITIALS TO AN *******")
        depDB["ANNOTATOR_INITIALS"] = "unknown"
    if "author_release_status_code" not in dataDict:
        depDB["AUTHOR_RELEASE_STATUS_CODE"] = "HPUB"
    else:
        depDB["AUTHOR_RELEASE_STATUS_CODE"] = dataDict["author_release_status_code"]
    depDB["TITLE"] = dataDict["title"].replace("'", " ")

    if "audit_author" in dataDict:
        auth = dataDict["audit_author"]
    else:
        auth = "unknown"

    if auth is not None:
        if isinstance(auth, list):
            depDB["AUTHOR_LIST"] = ""
            depDB["AUTHOR_LIST"] = " ".join(["%s," % a for a in auth])
        elif isinstance(auth, str):
            depDB["AUTHOR_LIST"] = auth
        else:
            logger.info(" Unknown type of author list %s", str(type(auth)))
        depDB["AUTHOR_LIST"] = depDB["AUTHOR_LIST"].replace("'", " ")
        logger.info(depDB["AUTHOR_LIST"])
    else:
        depDB["AUTHOR_LIST"] = "unknown"
    #
    depDB["EXP_METHOD"] = dataDict["exp_method"]
    if depDB["EXP_METHOD"].find("NMR") >= 0:
        if "status_code_nmr" in dataDict:
            depDB["STATUS_CODE_EXP"] = dataDict["status_code_nmr"]
        else:
            depDB["STATUS_CODE_EXP"] = "Unknown"
    elif depDB["EXP_METHOD"].find("EM") >= 0:
        if "status_code_em" in dataDict:
            depDB["STATUS_CODE_EXP"] = dataDict["status_code_em"]
        else:
            depDB["STATUS_CODE_EXP"] = "Unknown"
    else:
        if "status_code_sf" in dataDict:
            depDB["STATUS_CODE_EXP"] = dataDict["status_code_sf"]
        else:
            depDB["STATUS_CODE_EXP"] = "Unknown"
    if "full_name_of_center" in dataDict:
        depDB["SG_CENTER"] = dataDict["full_name_of_center"]
    else:
        depDB["SG_CENTER"] = "Unknown"

    if DBstatusAPI.exist(depDB):
        # then we do nothing - since this workflow does not modify the depostion
        constDict = {}
        constDict["DEP_SET_ID"] = depositionID
        DBstatusAPI.saveObject(depDB, "update", constDict)
    else:
        DBstatusAPI.saveObject(depDB, "insert")
        if debug > 0:
            prt.write("WFE.WFEapplications.initialliseDB : done insert data object in DB\n")
    #
    classDB = {}
    classDB["WF_CLASS_ID"] = WorkflowClassID
    classDB["WF_CLASS_FILE"] = metaData.getName()
    classDB["WF_CLASS_NAME"] = metaData.getName()
    classDB["TITLE"] = metaData.getDescription()
    classDB["AUTHOR"] = metaData.getAuthor()
    classDB["VERSION"] = metaData.getVersionMajor() + metaData.getVersionMinor()

    if DBstatusAPI.exist(classDB):
        if debug > 0:
            prt.write("WFE.WFEapplications.initialliseDB : updating the workflow class\n")
        # do nothing as we have run this workflow
        # We need to review the version to see if it is the same - how do we
        #   handle a new version of the workflow-class ?
        constDict = {}
        constDict["WF_CLASS_ID"] = WorkflowClassID
        DBstatusAPI.saveObject(classDB, "update", constDict)
    else:
        if debug > 0:
            prt.write("WFE.WFEapplications.initialliseDB : creating the workflow class object\n")
        DBstatusAPI.saveObject(classDB, "insert")

    instDB = {}
    instDB["WF_INST_ID"] = WorkflowInstanceID
    instDB["WF_CLASS_ID"] = WorkflowClassID
    instDB["DEP_SET_ID"] = depositionID
    instDB["STATUS_TIMESTAMP"] = datetime.datetime.utcnow()
    DBstatusAPI.updateStatus(instDB, "init")

    # and also initialise the wf_instance_last table
    sql = "update wf_instance_last set status_timestamp=" + str(getTimeNow()) + ", inst_status='init' where dep_set_id = '" + depositionID + "'"
    logger.info(sql)
    ok = DBstatusAPI.runUpdateSQL(sql)
    if ok < 1:
        logger.info(" CRITICAL : failed to update wf_instance_last table")

    if log != sys.stderr:
        log.close()


# XXXXX Believe internal to this file only - privatize scope?
def updateDeposit(DBstatusAPI, depositionID, site="pdbe", title="?", auth="?", method="?"):
    """
    update deposition communications table
    """

    depDB = {}
    depDB["DEP_SET_ID"] = depositionID
    timeNow = getTimeNow()

    if DBstatusAPI.exist(depDB):

        logger.info(" -> Running SQL 1 : update deposition")
        if title is None:
            sql = "update deposition set annotator_initials = 'unknown',status_code = 'PROC'   where dep_set_id = '%s' and status_code = 'DEP'" % (depositionID)
        else:
            sql = (
                "update deposition set annotator_initials = 'unknown',status_code = 'PROC', deposit_site = '%s', title = '%s', author_list = '%s', exp_method = '%s' where dep_set_id = '%s' and status_code = 'DEP'"  # noqa: E501
                % (site, title, auth, method, depositionID)
            )
        logger.info(sql)
        nrow = DBstatusAPI.runUpdateSQL(sql)
        logger.info("Rows affected: %i", nrow)

        logger.info(" -> Running SQL 2 : insert wf_instance")
        # make sure we only get the last in case processing was done with new instance data
        sql = "select ordinal from wf_instance where dep_set_id = '" + str(depositionID) + "'  order by status_timestamp desc limit 1"
        allList = DBstatusAPI.runSelectSQL(sql)

        if allList is not None:
            for allRow in allList:
                ordinal = allRow[0]
                sql = (
                    "update wf_instance set wf_inst_id = 'W_001', wf_class_id = 'Annotate', owner = 'Annotation.bf.xml', inst_status = 'init', status_timestamp = '%s' where ordinal = '%s'"
                    % (timeNow, str(ordinal))
                )
                logger.info(sql)
                nrow = DBstatusAPI.runInsertSQL(sql)
                logger.info("Rows affected: %i", nrow)
                break

        logger.info(" -> Running SQL 3 : insert wf_instance_last")
        sql = (
            "update wf_instance_last set wf_inst_id = 'W_001', wf_class_id = 'Annotate', owner = 'Annotation.bf.xml', inst_status = 'init', status_timestamp = '%s' where dep_set_id = '%s'"
            % (timeNow, str(depositionID))
        )
        logger.info(sql)
        nrow = DBstatusAPI.runInsertSQL(sql)
        logger.info("Rows affected: %i", nrow)

        return nrow

    return 0


# Deprecated - but need to cleanup WF examples
def initilliseDeposit(
    DBstatusAPI,
    depositionID,
    pdb="?",
    date=None,  # pylint: disable=unused-argument
    initials="unknown",  # pylint: disable=unused-argument
    deposit_site="?",
    process_site="?",
    status_code="DEP",  # pylint: disable=unused-argument
    author_code="?",
    title="?",
    author_list="?",
    expt="?",
    status_code_exp="?",
    SG_center="?",
    ann="dep",
    email="",
):
    """
    raw deposition table inialisation
    """

    depDB = {}
    depDB["DEP_SET_ID"] = depositionID

    if DBstatusAPI.exist(depDB):
        logger.info("ID exists ")
        sql = "update deposition set process_site = '" + str(process_site) + "', exp_method = '" + str(expt) + "' where dep_set_id = '" + str(depositionID) + "'"
        logger.info(sql)
        DBstatusAPI.runUpdateSQL(sql)
    else:
        sql = (
            "insert deposition (dep_set_id, pdb_id, deposit_site, process_site, status_code, author_release_status_code, title, author_list, exp_method, status_code_exp, SG_center, annotator_initials,email) values ('"  # noqa: E501
            + str(depositionID)
            + "','"
            + str(pdb)
            + "','"
            + str(deposit_site)
            + "','"
            + str(process_site)
            + "','"
            + str(status_code)
            + "','"
            + str(author_code)
            + "','"
            + str(title)
            + "','"
            + str(author_list)
            + "','"
            + str(expt)
            + "','"
            + str(status_code_exp)
            + "','"
            + str(SG_center)
            + "','"
            + str(ann)
            + "','"
            + str(email)
            + "')"
        )
        logger.info(sql)
        DBstatusAPI.runInsertSQL(sql)


# Only used by examples XXX
def wfLogDirectory(depID):
    siteId = getSiteId(defaultSiteId="WWPDB_DEPLOY_TEST")
    cI = ConfigInfo(siteId)

    topSessionPath = cI.get("SITE_WEB_APPS_TOP_SESSIONS_PATH")
    logDir = os.path.join(topSessionPath, "wf-logs", depID)

    if not os.path.exists(logDir):
        os.makedirs(logDir)

    return logDir



##
# File:    CommandLineArgs.py
# Date:    15-Mar-2009
#
# Updates:
#  22-April-2010 : Incorporation into enterpise.
#
#  2-May-2015   jdw jdw remove -gui/interface arg.
#               jdw refacotor
##

"""
Interpret the run time parameters of the WFE
"""

__docformat__ = "restructuredtext en"
__author__ = "Tom Oldfield"
__email__ = "oldfield@ebi.ac.uk"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.24"

import sys
import getopt


class CommandLineArgs(object):

    """
    run time parameter class
    """

    def __init__(self, argv):
        self.workflow = "Workflow-initial-V2.xml"
        self.path = "./"
        self.sessionID = ""
        self.initTask = "entry-point"
        self.debug = 1
        self.processState = "PROC"
        self.instanceID = None
        self.accession = "1UIS"
        self.log = None
        self.wfDepth = 0  # which WF file name to get back from a nested name

        try:
            opts, _args = getopt.getopt(
                argv[1:],
                "hd:l:s:t:i:w:p:a:k:z:xr",
                ["help", "debug=", "log=", "session=", "task=", "instance=", "workflow=", "path=", "processState=", "accession=", "depth=", "xample", "recover"],
            )
        except getopt.GetoptError:
            # print help information and exit:
            self.usage()
            return None

        for opt, arg in opts:
            #      print  " opt " + opt + " : arg " + arg
            if opt in ("-h", "--help"):
                self.usage()
                return None
            elif opt in ("-d", "--debug"):
                self.debug = int(arg)
            elif opt in ("-l", "--log"):
                self.log = arg
            elif opt in ("-t", "--task"):
                self.initTask = arg
            elif opt in ("-i", "--instance"):
                self.instanceID = arg
            elif opt in ("-s", "--session"):
                self.sessionID = arg
            elif opt in ("-p", "--path"):
                self.path = arg
            elif opt in ("-w", "--workflow"):
                self.workflow = arg
            elif opt in ("-a", "--accession"):
                self.accession = arg
            elif opt in ("-k", "--processState"):
                self.processState = arg
            elif opt in ("-z", "--depth"):
                self.wfDepth = int(arg)
            elif opt in ("-r", "--recover"):
                self.processState = arg

    def getDebug(self):
        return self.debug

    def getSessionID(self):
        return self.sessionID

    def getProcessState(self):
        return self.processState

    def getPath(self):
        return self.path

    def getWFdepth(self):
        return self.wfDepth

    def getWorkFlowFileName(self):
        """
        -w annotation(sequenceModule)
        -w one(two(three(four)))
        """

        anyB = self.workflow.find("(")

        if anyB < 0 and self.wfDepth > 0:
            # systax error in wf request
            print("Request for workflow of depth " + str(self.wfDepth) + " that is invalid in " + self.workflow + " : zero level error")
            sys.exit(0)

        if anyB < 0:
            return self.workflow
        else:
            start = 0
            for i in range(0, self.wfDepth + 1):
                b1 = self.workflow.find("(", start)
                if b1 < 0:
                    print("Request for workflow of depth " + str(self.wfDepth) + " that is invalid in " + self.workflow + " : " + str(i) + " level error")
                    exit(0)
                if i == self.wfDepth:
                    return self.workflow[start:b1]
                start = b1 + 1

        return self.workflow

    def getInitTask(self):
        return self.initTask

    def usage(self):

        print("Usage of the engine")
        print("python mainEngine.py -s <session name> [-i <Task>] [-d 1]")
        print("-s  : Session name or deposition ID : required to run engine on domain data")
        print("-w  : workflow File name : no default")
        print("-p  : path for XML workflows : default = ./")
        print("-i  : instance ID")
        print("-g  : interface console (default - interface is external)")
        print("-t  : start workflow at this task")
        print("-d  : debug level 0/1/2/3 (default = 0)")
        print("-l  : log output file")
        print("-a  : accession code only for  -x")
        print("-x  : Insert dummy dep_id and class_id into DB")
        print("-r  : automatically recover instance from last working task")
        print("-k  : set domain data processState - only valid with -x")
        print("-z  : set the WFdepth (default = 0)")



##
# File:    mainEngine.py
# Date:    15-Mar-2009
#
# Updates:
#  22-April-2010 : Incorporation into enterpise.
#  09-August-2010 : Include logging
#  13-Feb-2015 jdw : major overhaul -
#  25-Apr-2015 jdw : another refactoring & continued revision of task processing -
#  2-May-2015  jdw : rewrote switch/branching task
#  2-Sep-2015  jdw : replace incomplete list of hardcoded storage types with api call --
#  7-Spe-2015  jdw : handle constant values assigned by api
#
##

"""

Workflow engine - reads the workflow and starts at
a point in the workflow given by the state of the
depositionID.

Entry point = "run" though this is not a threaded process

The WFE uses the first task/ instance ID and depID to start
  at the right point
"recovery" mode will also start at the last valid point in the WF

The rest is just a infinite loop of task follow task.

"""

__docformat__ = "restructuredtext en"
__author__ = "Tom Oldfield"
__email__ = "oldfield@ebi.ac.uk"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.24"

import sys
import time
import os
import logging
from xml.dom import minidom

# from threading import *

from wwpdb.apps.wf_engine.engine.SwitchTask import SwitchTask
from wwpdb.apps.wf_engine.engine.ExternalTask import ExternalTask
from wwpdb.apps.wf_engine.engine.WorkflowSession import WorkflowSession
from wwpdb.apps.wf_engine.engine.CommandLineArgs import CommandLineArgs
from wwpdb.apps.wf_engine.engine.ProcessManager import ProcessManager
from wwpdb.apps.wf_engine.engine.WorkflowManager import WorkflowManager

from wwpdb.apps.wf_engine.engine.InterpretDataObject import fillAPIinputObject
from wwpdb.apps.wf_engine.engine.InterpretDataObject import fillAPIoutputObject
from wwpdb.apps.wf_engine.engine.InterpretDataObject import getObjectValue
from wwpdb.apps.wf_engine.engine.InterpretDataObject import getTaskParameterDict


from wwpdb.apps.wf_engine.engine.EngineUtils import EngineUtils

#
from wwpdb.apps.wf_engine.reader.parseXML import parseXML

#
from wwpdb.apps.wf_engine.wf_engine_utils.time.TimeStamp import TimeStamp

from wwpdb.apps.wf_engine.wf_engine_utils.run.MyLogger import MyLogger
from wwpdb.apps.wf_engine.wf_engine_utils.process.ProcessUtils import ProcessUtils
from wwpdb.io.locator.DataReference import DataFileReference

logger = logging.getLogger(name="root")


class mainEngine(object):
    def __init__(self, debug=4, prt=None):  # pylint: disable=unused-argument
        logger.info("+mainEngine.__init__ Starts log level at %d", logger.getEffectiveLevel())
        # Global variables
        # path - this is a list that contains the tasks visited in order
        self.path = []
        # debug level
        self.debug = debug
        #        self.__lfh = MyLogger(level=logging.DEBUG)
        self.__lfh = MyLogger(level=logging.INFO)

        self.__verbose = debug > 0

        self.__depositionId = ""
        self.__wfClassId = ""
        self.__wfInstId = ""
        self.__wfClassFileName = ""

        # this should be available everywhere
        self.runTimeParameters = None

        # ---  Database connection and methods encapsulated in the EngineUtils() class
        self.__eUtil = EngineUtils(verbose=self.__verbose)
        self.__timeStamp = TimeStamp()
        #
        # Something went wrong - exception task
        self.exception = None
        self.instanceExitState = "finished"

    def __createDBclass(self, wfMetaDataObj):
        """
        Internal method create a workflow-class reference if not present.
        The statusDB class-dict contains one reference to each workflow class - so we only update when we find
        a new one.
        """

        classDB = {}
        classDB["WF_CLASS_ID"] = wfMetaDataObj.getID()
        classDB["WF_CLASS_FILE"] = wfMetaDataObj.getName()
        classDB["WF_CLASS_NAME"] = wfMetaDataObj.getName()
        classDB["TITLE"] = wfMetaDataObj.getDescription()
        classDB["AUTHOR"] = wfMetaDataObj.getAuthor()
        classDB["VERSION"] = wfMetaDataObj.getVersionMajor() + wfMetaDataObj.getVersionMinor()

        db = self.__eUtil
        if db.exist(classDB):
            if self.debug > 0:
                logger.info("+mainEngine.__createDBclass : found existing instance for workflow class %s file %s", wfMetaDataObj.getID(), wfMetaDataObj.getName())
        else:
            if self.debug > 0:
                logger.info("+mainEngine.__createDBclass : inserting new workflow class object for %s file %s", wfMetaDataObj.getID(), wfMetaDataObj.getName())
            db.saveObject(classDB, "insert")

        if self.debug > 1:
            logger.info("+mainEngine.__createDBclass :  done with classDB")

    def __setDBInstStatus(self, typein, mode, task):  # pylint: disable=unused-argument
        """
        Method to update the instance data in the status DB
        The WFM uses this instance data to reduce contention.

        The method always does an update since the instance
        data must exist on entry to the workflow - and a new
        workflow instance is always created with each run
        """

        now = self.__timeStamp.getSecondsFromReference()
        instDB = {}
        instDB["WF_INST_ID"] = self.__wfInstId
        instDB["WF_CLASS_ID"] = self.__wfClassId
        instDB["DEP_SET_ID"] = self.__depositionId
        #    instDB['STATUS_TIMESTAMP'] = datetime.datetime.utcnow()
        instDB["STATUS_TIMESTAMP"] = now

        self.__eUtil.updateConnection()
        db = self.__eUtil
        db.updateStatus(instDB, mode)

        # Tom :  update the wf_instance_last table - note it is unique over dep_set_id

        sql = (
            "update wf_instance_last set status_timestamp="
            + str(now)
            + ", inst_status='"
            + mode
            + "', wf_class_id = '"
            + self.__wfClassId
            + "', wf_inst_id = '"
            + self.__wfInstId
            + "' where dep_set_id = '"
            + self.__depositionId
            + "'"
        )
        ok = self.__eUtil.runUpdateSQL(sql)
        if ok < 1:
            logger.error("+mainEngine.__setDBInstStatus - failed to update wf_instance_last table for %s", self.__depositionId)

    def __setDBTaskStatus(self, typein, mode, task):
        """
        Method to update the task data in the status DB
        Current :
          The exist status of the task must be done since
          a loop in the workflow might mean we revisit a task
        To be done and supported by sytem :
          A status should always insert with a new task
          since we need to record all loop structure in WF
        """

        taskID = {}
        taskID["WF_TASK_ID"] = task.name
        taskID["WF_INST_ID"] = self.__wfInstId
        taskID["WF_CLASS_ID"] = self.__wfClassId
        taskID["DEP_SET_ID"] = self.__depositionId
        taskID["TASK_NAME"] = task.name
        taskID["TASK_TYPE"] = typein
        taskID["TASK_STATUS"] = mode
        #    taskID['STATUS_TIMESTAMP'] = datetime.datetime.utcnow()
        taskID["STATUS_TIMESTAMP"] = self.__timeStamp.getSecondsFromReference()

        # This version only creates unique taskID
        if self.__eUtil.exist(taskID):
            self.__eUtil.updateStatus(taskID, mode)
        else:
            self.__eUtil.saveObject(taskID, "insert")

    # def __getNewInstanceIDDB(self, depositionId):
    #     """
    #       Create new workflow instance from the statusDB
    #       Returns the ID
    #       The status API returns an ID as a number, this method
    #       prefixes the numeric with a W_
    #     """

    #     # This returns unique instID on depID AND the classID ; want the next one on depID only
    #     #    instID = self.__eUtil.getNextWfInstId(self.__depositionId,self.__wfClassId)

    #     sql = "select max(cast(substr(wf_inst_id,3) as decimal(5,0)))+1 from wf_instance where dep_set_id = '" + str(depositionId) + "'"

    #     ll = self.__eUtil.runSelectSQL(sql)

    #     if ll is None:
    #         instID = 1
    #     else:
    #         for l in ll:
    #             if l is None:
    #                 instID = 1
    #             else:
    #                 if l[0] is None:
    #                     instID = 1
    #                 else:
    #                     instID = l[0]

    #     ret = "W_" + str(instID).zfill(3)

    #     if self.debug > 0:
    #         logger.info("+mainEngine.__getNewInstanceIDDB : for data set ID = %s next workflow instance ID = %s", depositionId, ret)

    #     return ret

    def __createNewDBwfInstance(self):
        """
        Creates a new instance record within the status DB.
        A duplicate should NEVER be created and is a problem
        of the API-instance ID generator
        """

        instDB = {}

        #        self.__wfInstId = self.__getNewInstanceIDDB(self.__depositionId)
        self.__wfInstId = self.__eUtil.getNextInstanceId(self.__depositionId)

        now = self.__timeStamp.getSecondsFromReference()
        instDB["WF_INST_ID"] = self.__wfInstId
        instDB["WF_CLASS_ID"] = self.__wfClassId
        instDB["DEP_SET_ID"] = self.__depositionId
        instDB["OWNER"] = self.__wfClassFileName
        instDB["INST_STATUS"] = "running"
        instDB["STATUS_TIMESTAMP"] = now
        if not self.__eUtil.exist(instDB):
            self.__eUtil.saveObject(instDB, "insert")
        else:
            logger.info("+mainEngine.__createNewDBWfInstance  *** Error : duplicate instance %s", str(self.__wfInstId))
            constDict = {}
            constDict["WF_INST_ID"] = self.__wfInstId
            self.__eUtil.saveObject(instDB, "update", constDict)

        # Tom New code for the wf_instance_last table

        sql = "select ordinal from wf_instance_last where dep_set_id = '" + self.__depositionId + "'"
        elist = self.__eUtil.runSelectSQL(sql)

        if elist is None or len(elist) == 0:
            ind = -1
        else:
            for el in elist:
                if el is None:
                    ind = -1
                else:
                    ind = el[0]

        if ind < 0:
            # Create new row
            sql = (
                "insert wf_instance_last(wf_inst_id,wf_class_id,dep_set_id,owner,inst_status,status_timestamp) values ('"
                + self.__wfInstId
                + "','"
                + self.__wfClassId
                + "','"
                + self.__depositionId
                + "','"
                + self.__wfClassFileName
                + "','running',"
                + str(now)
                + ")"
            )
            ok = self.__eUtil.runInsertSQL(sql)
        else:
            # sql = "update wf_instance_last set wf_inst_id='" +
            # self.__wfInstId + "', wf_class_id='" + self.__wfClassId +
            # "', dep_set_id='" +  self.__depositionId + "',  inst_status='running',
            # status_timestamp=" + str(now) + " where ordinal = " + str(ind)
            sql = (
                "update wf_instance_last set wf_inst_id='"
                + self.__wfInstId
                + "', wf_class_id='"
                + self.__wfClassId
                + "', dep_set_id='"
                + self.__depositionId
                + "', owner='"
                + self.__wfClassFileName
                + "', inst_status='running', status_timestamp="
                + str(now)
                + " where ordinal = "
                + str(ind)
            )
            ok = self.__eUtil.runUpdateSQL(sql)

        if ok < 1:
            logger.error("+mainEngine.__createNewDBWfInstance failed to update wf_instance_last table")
        else:
            logger.info("+mainEngine.__createNewDBWfInstance wf_instance[_last] tables updated with %s", str(instDB))

    # def __doExceptionTask(self):
    #     """
    #       Global exception handler - this is the unhandled
    #         exception only
    #       Need to review this.
    #     """
    #     # for now just exit(1) - but we need to find the next task

    #     logger.error("+mainEngine.__doExceptionTask.  : called exit ")
    #     sys.exit(1)

    def __findEntryPoint(self, taskObjList):
        """
         find the starting point in the workflow
        - this is the default starting point
        """

        startingPoint = self.__findTaskByType("Entry-point", taskObjList)

        return startingPoint

    def __findTaskByType(self, typein, taskObjList):
        """
        Returns a task found by the task type, only of use
        for start/stop/exception - since the return will be
        ambigious for other types
        returns the task object
        """

        for task in taskObjList:

            if task.type.lower() == typein.lower():
                if self.debug > 1:
                    logger.info("+mainEngine.findTaskByType : Found task with type = %s", str(task.type))

                return task

        logger.error("+mainEngine.findTaskByType : Failed to find a task-by-type %s", str(typein))

        sys.exit(0)

    def __findTaskNameFromHumanName(self, humanName, taskObjList):
        """
        Returns a task found by the task name(ie unique ID) returns the task object
        """

        for task in taskObjList:
            if task.nameHumanReadable == humanName:
                if self.debug > 1:
                    logger.info("+mainEngine.__findTaskNameFromNumanName :   Found task with name = %s", str(task.name))
                return task

        # this method should not die
        return None

    def __findTaskByName(self, name, taskObjList):
        """
        Returns a task found by the task name(ie unique ID)
        returns the task object
        """

        for task in taskObjList:
            if task.name == name:
                if self.debug > 1:
                    logger.info("+mainEngine.__findTaskByName :   Found named task %s", str(task.name))
                return task

        logger.error("+mainEngine.__findTaskByName - Failed to find a task-by-name %s", str(name))
        return None

    def __loopTest(self, task, iterator, value):
        """
        method to manage a loop task
        It is really a test against an iterator.  Notice that the loop
        test only returns 2 states, the normal next task, or the exit
        loop next-task.  The usable return state is that the loop
        variable data object is made equal to the i'th iterator value.
        1) if the loop variable is not set - set it to the first iterator
        2) if the loop variable is set, then test it against the iterator
           list and set the loop variable = next item : return 0 :
           the first "nextTask"
        3) If we are at the end of the loop then return 1 : the second "nextTask"
        """

        for _key, data in value.items():
            # valueKey = key
            valueData = getObjectValue(data, self.debug, self.__lfh)
        for _key, data in iterator.items():
            # iteratorKey = key
            iteratorData = getObjectValue(data, self.debug, self.__lfh)

        if not isinstance(iteratorData, list):
            logger.info("+mainEngine.__loopTest : *** Error in WF - can only use a list for loop %s", str(type(iteratorData)))
            logger.info("+mainEngine.__loopTest : *** Error in WF - This is the data %s", str(iteratorData))
            return -1

        if self.debug > 1:
            logger.info("+mainEngine.__loopTest :   Loop output variable %s", str(valueData))
            logger.info("+mainEngine.__loopTest :   Loop input iterator %s", str(iteratorData))
            logger.info("+mainEngine.__loopTest :   Loop input length %s", str(len(iteratorData)))

        if self.debug > 0:
            logger.info(" +mainEngine.__loopTest : loop at %s / %s", str(valueData), str(len(iteratorData)))

        if valueData is None:
            # we are starting the loop
            if self.debug > 1:
                logger.info("+mainEngine.__loopTest : ValueData  =  %s", str(valueData))
                logger.info("+mainEngine.__loopTest : task.param =  %s", str(task.param))
            for _key, data in value.items():
                if len(iteratorData) < 1:
                    if self.debug > 1:
                        logger.info("+mainEngine.__loopTest :   Loop list of zero length \n")
                    return 1
                if task.param is not None and task.param == "index":
                    data.data = str("1")
                else:
                    data.data = iteratorData[0]
                if self.debug > 1:
                    logger.info("+mainEngine.__loopTest :  starting loop %s", str(data.data))
                return 0
        else:
            if task.param is not None and task.param == "index":
                index = int(valueData)
                if index >= len(iteratorData) - 1:
                    if self.debug > 1:
                        logger.info("+mainEngine.__loopTest : (index return) found end of loop ")
                    return 1
                else:
                    for _key, data in value.items():
                        data.data = str(index + 1)
                        if self.debug > 1:
                            logger.info("+mainEngine.__loopTest (index return) - Data value : %s", str(data.data))
                        return 0
            else:
                for index, item in enumerate(iteratorData):
                    if item == valueData:
                        if index >= len(iteratorData) - 1:
                            if self.debug > 1:
                                logger.info("+mainEngine.__loopTest : (value return) found end of loop ")
                            return 1
                        else:
                            for _key, data in value.items():
                                data.data = iteratorData[index + 1]
                                if self.debug > 1:
                                    logger.info("+mainEngine.__loopTest :  (value return) data value %s", str(data.data))
                                return 0

        return -1

    def __runProcess(self, task, dataValues, dataObjList):  # pylint: disable=unused-argument
        """
        JDW  Revised method to run single process task in a thread -

        """
        logger.info("+mainEngine.__runProcess :  Starting process %s", task.name)
        self.__setDBTaskStatus("process", "init", task)

        # setup the process
        self.__setDBTaskStatus("process", "start", task)
        process = ProcessManager(task, self.debug)
        process.setStatsInfo(self.__depositionId, self.__wfClassId, self.__wfInstId)
        dataForTask = task.getDataReference(dataObjList)
        if dataForTask:
            dataInput = self.__handleIOData("input", task, dataForTask, dataObjList)
            taskParamD = getTaskParameterDict(dataObjList, task.parameter, self.debug, self.__lfh)
            process.setInput(dataInput, taskParamD)
            dataOutput = self.__handleIOData("output", task, dataForTask, dataObjList)
            process.setOutput(dataOutput)

        # start the process
        process.start()
        processThread = process

        status = "running"
        #   find the longest failtime of all processes
        failTime = 0
        if task.failTime == 0:
            failTime = 2147483640  # about 68 years
        if task.failTime > failTime:
            failTime = task.failTime

        loop = 0
        while loop < failTime:
            loop = loop + 1
            time.sleep(1)

            # check if any crashed
            status = processThread.getState()
            if self.debug > 2:
                logger.debug("+mainEngine.__runProcess :  current status = %s", str(status))
            if self.__isException(status):
                logger.info("+mainEngine.__runProcess :  Task %s crashed", task.name)
                self.__setDBTaskStatus("process", "crashedX", task)
                self.__setDBInstStatus("process", "exception", task)
                self.exception = status
                return "exception"

            #   have they all completed ?
            status = processThread.getState()
            if status == "complete":
                logger.info("+mainEngine.__runProcess : task complete : %s", task.name)
                self.__setDBTaskStatus("process", "complete", task)
                self.instanceExitState = processThread.getExitState()
                return "complete"

        # The fail time is reached : abort any running tasks
        self.__setDBTaskStatus("process", "timeoutX", task)
        processThread.abort()
        logger.error("+mainEngine.__runProcess :  Trying to kill process")

        pid = os.getpid()
        logger.info("+mainEngine.__runProcess Stopping children %s id %s", str(processThread.getName()), str(pid))
        pu = ProcessUtils(verbose=True, log=self.__lfh)
        cPidList = pu.getChildren(pid)
        myStatus = pu.killProcessList(cPidList)
        logger.info("+mainEngine.__runProcess : Child kill status is %r", myStatus)
        time.sleep(1)
        #  record the task in exception and continue to dump the rest
        self.exception = "timeoutX"
        self.__setDBInstStatus("process", "exception", task)
        return "exception"

    def __runWorkflow(self, task, recoveryFlag):
        """
        Internal method to manage a workflow in a separate thread -
        """
        logger.info("+mainEngine.__runWorkflow() : %s  worflow %s instance %s recoveryFlag %r", self.__depositionId, task.name, self.__wfInstId, recoveryFlag)

        if recoveryFlag == 0:
            workflowThread = WorkflowManager(task.file, self.__depositionId, self.__wfInstId, "entry-point")
        elif recoveryFlag == 1:
            workflowThread = WorkflowManager(task.file, self.__depositionId, self.__wfInstId, "recover")
        elif recoveryFlag == 2:
            workflowThread = WorkflowManager(task.file, self.__depositionId, "recover", "recover")
        elif recoveryFlag == 3:
            # not used
            workflowThread = WorkflowManager(task.file, self.__depositionId, "Annotate", str(task.name))

        self.__setDBTaskStatus("workflow", "start", task)
        workflowThread.start()
        self.__setDBTaskStatus("workflow", "running", task)

        status = -1
        #
        failTime = 0
        if task.failTime == 0:
            failTime = 2147483640
        if task.failTime > failTime:
            failTime = task.failTime

        logger.info("+mainEngine.__runWorkflow() : %s  worflow %s timeout %d", self.__depositionId, task.name, failTime)
        loop = 0
        while loop < failTime:
            loop = loop + 1
            time.sleep(1)

            # check if any crashed
            status = workflowThread.workflowStatus()
            if status == 3:
                logger.info("+mainEngine.__runWorkflow() : %s  workflow %s crashed", self.__depositionId, task.name)
                self.__setDBTaskStatus("workflow", "crashedX", task)
                self.__setDBInstStatus("workflow", "exception", task)
                self.exception = "workflowX"
                return "exception"

            #   have they all completed ?
            # if status is 2 then all workflows have completed
            status = workflowThread.workflowStatus()
            if status == 2:
                logger.info("+mainEngine.__runWorkflow() : %s  workflow %s completed", self.__depositionId, task.name)
                self.__setDBTaskStatus("workflow", "complete", task)
                return "complete"

            # check that all started
            status = workflowThread.workflowStatus()
            if status == 0:
                logger.info("+mainEngine.__runWorkflow() : %s  workflow %s not started (exception)", self.__depositionId, task.name)
                self.__setDBTaskStatus("workflow", "startX", task)
                self.__setDBInstStatus("workflow", "exception", task)
                return "exception"

        # The fail time is reached : abort any running tasks
        self.__setDBTaskStatus("workflow", "timeoutX", task)
        workflowThread.abort()
        time.sleep(1)
        logger.info("+mainEngine.__runWorkflow() : %s  workflow %s timed out (exception)", self.__depositionId, task.name)
        self.__setDBInstStatus("workflow", "exception", task)
        return "exception"

    def __handleTask(self, task, dataObjList, recoveryFlag):
        """
        Input: - task is list of tasks objects (LENGTH=1 always)

        Takes control of a task and runs the task
        1) checks whether there a valid parallel tasks
        2) parallel processes and workflows are run elsewhere
        3) get the data for the single workflow path
        4) run the relevant task exception manager + task
        """
        logger.info(
            "+mainEngine.__handleTask : -------------------------------------------   STARTS for class %s TASK %s   -------------------------------------- ",
            self.__wfClassId,
            task.name,
        )
        logger.info(
            "+mainEngine.__handleTask : starting for %s class %s task %s  type %s recoveryFlag %r", self.__depositionId, self.__wfClassId, task.name, task.type, recoveryFlag
        )

        if task.type != "Manual":
            self.__setDBTaskStatus(task.type, "init", task)

        ret = 0
        status = "init"
        # real data dictionary of data values used for input and output
        dataValues = {}

        if task.type == "Workflow":
            logger.info("+mainEngine.__handleTask : %s calling runWorkflow for class %s task %s recoveryFlag %r", self.__depositionId, self.__wfClassId, task.name, recoveryFlag)
            status = self.__runWorkflow(task, recoveryFlag)
            ret = 0

        elif task.type == "Process":
            logger.info("+mainEngine.__handleTask : %s calling runProcess for class %s task %s", self.__depositionId, self.__wfClassId, task.name)
            status = self.__runProcess(task, dataValues, dataObjList)
            ret = 0

        else:
            logger.info("+mainEngine.__handleTask : %s starting inline handler for class %s task %s", self.__depositionId, self.__wfClassId, task.name)
            task.dump(self.__lfh)
            dataForTask = task.getDataReference(dataObjList)

            inputValues = None
            outputValues = None
            if dataForTask:
                inputValues = self.__handleIOData("input", task, dataForTask, dataObjList)
                outputValues = self.__handleIOData("output", task, dataForTask, dataObjList)

            if self.debug > 1:
                if dataForTask:
                    logger.info("+mainEngine.__handleTask :  Data objects for in-line task %s type %s", task.name, task.type)
                    for dat in dataForTask:
                        dat.printMe(self.__lfh)

            if task.type == "Manual":
                logger.info(
                    "+mainEngine.__handleTask :  %s EXTERNAL MANUAL task for class %s task %s with recoveryFlag %r", self.__depositionId, self.__wfClassId, task.name, recoveryFlag
                )
                if recoveryFlag == 0:
                    self.__setDBTaskStatus(task.type, "start", task)
                manual = ExternalTask(self.__eUtil, self.__depositionId, self.__wfClassId, self.__wfInstId, task, self.debug, recoveryFlag)
                # need to get the correct name of the data object
                ret = manual.handleTask()
                if ret < 0:
                    self.__setDBTaskStatus(task.type, "exception", task)
                    self.__setDBInstStatus(task.type, "exception", task)
                    logger.info("+mainEngine.__handleTask set TASK/INSTANCE status for EXTERNAL MANUAL task %r to exception", task.name)
                    self.exception = "manualX"
                    status = "exception"

                else:
                    self.__eUtil.addReferenceOverwrite(self.__depositionId, self.__wfClassId, None, None, task.name, str(ret))
                    logger.info("+mainEngine.__handleTask set TASK status for EXTERNAL MANUAL task %r to complete", task.name)
                    self.__setDBTaskStatus(task.type, "complete", task)
                    status = "complete"

            elif task.type == "Loop":
                logger.info("+mainEngine.__handleTask :  loop class %s task %s ", self.__wfClassId, task.name)
                self.__setDBTaskStatus(task.type, "start", task)
                ret = self.__loopTest(task, inputValues, outputValues)
                if ret < 0:
                    # then the function return resulted in exception
                    self.__setDBTaskStatus(task.type, "exception", task)
                    self.__setDBInstStatus(task.type, "exception", task)
                    self.exception = "loopX"
                    status = "exception"
                else:
                    status = "complete"
            elif task.type == "Decision":
                logger.info("+mainEngine.__handleTask :  %s switch class %s task %s type %s ", self.__depositionId, self.__wfClassId, task.name, task.type)
                self.__setDBTaskStatus(task.type, "start", task)
                switch = SwitchTask(task, debug=self.debug)
                # need to get the correct name of the data object
                self.__setDBTaskStatus(task.type, "running", task)
                ret = switch.handleTask(inputValues)
                # put the return value into the status database
                if ret < 0:
                    # then the function return resulted in exception
                    self.__setDBTaskStatus(task.type, "exception", task)
                    self.__setDBInstStatus(task.type, "exception", task)
                    self.exception = "decisionX"
                    status = "exception"
                else:
                    self.__eUtil.addReferenceOverwrite(self.__depositionId, self.__wfClassId, None, None, task.name, str(ret))
                    self.__setDBTaskStatus(task.type, "complete", task)
                    status == "complete"  # pylint: disable=pointless-statement
            elif task.type == "Entry-point":
                status == "complete"  # pylint: disable=pointless-statement
            elif task.type == "Exit-point":
                status == "complete"  # pylint: disable=pointless-statement
            elif task.type == "Exception":
                logger.info("+mainEngine.__handleTask :  Exception handler thown in workflow class %s - aborting", self.__wfClassId)
                self.__setDBTaskStatus(task.type, "abort", task)
                self.__setDBInstStatus(task.type, "abort", task)
                logger.info("+mainEngine.__handleTask : exiting for %s with status %r ret %r", self.__depositionId, status, ret)
                sys.exit(1)
            else:
                logger.info("+mainEngine.__handleTask : failure : class %s undefined task type %r ", self.__wfClassId, task.type)
                logger.info("+mainEngine.__handleTask : exiting for %s with status %r ret %r", self.__depositionId, status, ret)
                sys.exit(0)

        if not status == "exception":
            self.__setDBTaskStatus(task.type, "finished", task)

        logger.info("+mainEngine.__handleTask : returning for %s with status %r ret %r", self.__depositionId, status, ret)
        return ret

    def __handleIOData(self, mode, task, dataForTask, dataObjList):
        """
        Method to return data object dictionaries
        The data objects for the task as searched and cross referenced
          with the data in the data list

        if a processName has been defined then we have an API process
          which requires that we call objects by pre-defined names
        if a processName is undefined then we are using a switch task
          task or other non-API process.  Just pass data back as required
        The names of the API-process data objects are looked up and
          cross-referenced with the WF data objects.
        Where there is multiple input/output, then the object number
          is matched
        """

        # Need to handle the name of the data object properly
        # Warning - there is no sorted order to the data input names
        # how to we handle which item is which ????

        logger.info(
            "+mainEngine.__handleIOData : Starts for task %s mode %s action %s where %s data object length %d",
            task.name,
            mode,
            task.uniqueAction,
            task.uniqueWhere,
            len(dataForTask),
        )
        dfr = DataFileReference()
        storageTypeList = dfr.getStorageTypeList()
        #
        ret = {}
        nWFData = 0
        for data in dataForTask:
            if data.localMode.startswith(mode) or data.localMode.startswith("both"):
                if self.debug > 0:
                    logger.info("+mainEngine.__handleIOData :  Found %s reference %s", str(mode), str(data.name))
                    logger.info("+mainEngine.__handleIOData :  Storage type = %s", str(data.where))
                if (data.where in storageTypeList) or (data.where in ["constant", "inline"]) or (data.where[0:4] == "path"):
                    nWFData = nWFData + 1
                else:
                    pass
        #
        # Only if the processName is defined then we have a process
        # Other tasks will have no uniqueAction defined - so we
        # just get the data and call it by the uniqueID of the data object

        if task.uniqueAction is None:
            if self.debug > 0:
                logger.info("+mainEngine.__handleIOData :  Using task data from wfe")
        elif task.uniqueWhere == "wfe":
            if self.debug > 0:
                logger.info("+mainEngine.__handleIOData :  Using task data from wfe")
        elif task.uniqueWhere == "api":
            if self.debug > 0:
                logger.info("+mainEngine.__handleIOData :  Using task data from api")

            nProcessData = 0
            if mode == "input":
                nProcessData = task.ActionRegistry.getInputObjectCount(task.uniqueAction)
                ioNames = task.ActionRegistry.getInputObjectNames(task.uniqueAction)
                if self.debug > 1:
                    logger.info("+mainEngine.__handleIOData :  Number of input data = %s for fask name %s", str(nProcessData), str(task.uniqueAction))
                    logger.info("+mainEngine.__handleIOData :  registry input data names = %s", str(ioNames))
            elif mode == "output":
                nProcessData = task.ActionRegistry.getOutputObjectCount(task.uniqueAction)
                ioNames = task.ActionRegistry.getOutputObjectNames(task.uniqueAction)
                if self.debug > 1:
                    logger.info("+mainEngine.__handleIOData :  Number of output data = %s task name %s", str(nProcessData), str(task.uniqueAction))
                    logger.info("+mainEngine.__handleIOData :  registry output data names = %s", str(ioNames))
            else:
                logger.info("+mainEngine.__handleIOData :   Error : undefined IO mode in handleIOData %s", str(mode))

            if nProcessData == 0 and nWFData > 0:
                logger.info("+mainEngine.__handleIOData :  WARNING : input defined in WF - none required for process : %s IO = %s ", str(task.uniqueAction), str(mode))
                return None
            if nProcessData > 0 and nWFData == 0:
                logger.info("+mainEngine.__handleIOData :  WARNING : no input defined in WF - input required for process : %s IO = %s", str(task.uniqueAction), str(mode))
                return None
            if nProcessData < nWFData:
                logger.info("+mainEngine.__handleIOData :  WARNING : not enough data define in WF for process %s IO = %s", str(task.uniqueAction), str(mode))
                return None
            if nProcessData > nWFData:
                logger.info("+mainEngine.__handleIOData :  WARNING : too many data tems define in WF for process %s IO = %s", str(task.uniqueAction), str(mode))
                return None
        else:
            logger.info("+mainEngine.__handleIOData :   **** Unknown function <where> %s", str(task.uniqueWhere))

        nWFData = 0
        for data in dataForTask:
            logger.info(
                "+mainEngine.__handleIOData : name %r data.localMode %r data.where %r task.uniqueName %r task.uniqueWhere %r",
                data.name,
                data.localMode,
                data.where,
                task.uniqueName,
                task.uniqueWhere,
            )
            if data.localMode.startswith(mode) or data.localMode.startswith("both"):
                #
                #   change for to include 'constants' returned from api -
                if (data.where in storageTypeList) or (data.where[0:4] == "path") or ((data.where in ["constant"]) and (task.uniqueWhere == "api")):
                    #
                    if data.localMode.startswith("input") or data.localMode.startswith("both"):
                        fillAPIinputObject(dataObjList, data, self.__depositionId, self.debug, self.__lfh)
                    if data.localMode.startswith("output") or data.localMode.startswith("both"):
                        fillAPIoutputObject(dataObjList, data, self.__depositionId, self.debug, self.__lfh)

                    # if there is no task.uniqueAction - then call the object by the WF name
                    #
                    if task.uniqueName is None:
                        ret[data.name] = data.ApiData
                    elif task.uniqueWhere == "api":
                        logger.info("+mainEngine.__handleIOData : match ioNames using %r ", data.localMode)
                        if data.localMode[-1] in ["1", "2", "3", "4", "5", "6", "7", "8", "9"]:
                            for ioName in ioNames:
                                if ioName[-1] == data.localMode[-1]:
                                    ret[ioName] = data.ApiData
                                    logger.info("+mainEngine.__handleIOData :   found new indexed data name %s, %s %s", str(ioName), str(data.localMode), str(data.name))
                        else:
                            ret[ioNames[nWFData]] = data.ApiData
                    else:
                        ret[data.name] = data.ApiData
                    nWFData = nWFData + 1
                elif data.where == "workflow" or data.where == "WF":
                    # This is a workflow state model instance, probably in memory too.
                    # THIS IS A REAL DATA VALUE
                    #         ret[data.name] = data.data
                    ret[data.name] = data
                elif data.where in ["constant"]:
                    # this is a variable or a constant in the data declaration
                    ret[data.name] = data
                else:
                    logger.info("+mainEngine.__handleIOData :   *** Catastrophic Workflow failure : Unknown data origin %s", str(data.where))
                    sys.exit(0)

        return ret

    def __getNextTask(self, lastTask, opt, taskObjList):
        """
        Return the next task from the lasttask, for a decision task it will return the option number task
        """

        nextTask = None

        name = lastTask.getNextTaskName(opt)

        if name:
            nextTask = self.__findTaskByName(name, taskObjList)
            if self.debug > 1:
                logger.info("+mainEngine.__getNextTask :   Found next task: %s", str(nextTask.name))

        else:
            logger.info("+mainEngine.__getNextTask :  *** Catastrophic workflow error : failed to find next task")
            logger.info("+mainEngine.__getNextTask :  in %s number %s", str(lastTask), str(opt))

        return nextTask

    def __readWorkFlow(self, wfFilePath, debugLevel=0):
        """
        Get dom object for workflow class file -
        """
        xmldoc = minidom.parse(wfFilePath)
        parse = parseXML(debugLevel, self.__lfh)
        return parse, xmldoc

    def __readWorkFlowHeader(self, parse, xmldoc):
        """
        Get workflow metadata object --
        """
        wfMetaDataObj = parse.getMetaData(xmldoc)
        return wfMetaDataObj

    def __readWorkFlowBody(self, parse, xmldoc, depositionID, instanceID):
        """
        Return the task object and data object lists from the workflow definition
        """
        taskObjList = parse.getTaskObjects(xmldoc)
        dataObjList = parse.getDataObjects(depositionID, instanceID, xmldoc)
        return taskObjList, dataObjList

    def __outputPath(self):
        """Outputs to the terminal the pathway taken by the workflow"""

        logger.info("+mainEngine.outputPath :  Path taken\n")

        for p in self.path:
            logger.info(" -> %s", str(self.__prettyPrintTask(p)))
        logger.info("\n")

    def __prettyPrintTask(self, t):
        """
        Simple output method to print out the task name for humans
        """

        if t.name.lower() == t.nameHumanReadable.lower():
            return t.name
        else:
            return "(" + t.name + ":" + t.nameHumanReadable + ")"

    def __manageException(self, workingTask, opt, taskObjList):
        """
        General exception manager:
        if opt < 0 : Ouch - bad return state from task - basically
          no return task could be found = really a workflow class error
          instance is marked as exception

        if self.exception == exception : generic exception
           do not reset the exception state - instance will be marked as "exception"
        if self.exception == some other valid exception type (see below)

        if the exception task has a nextTask - this is used
        if the exception task has handles - then the self.exception is matched and used
        if the exception task has nothing - then it just exits(1)


        Exceptions
           exception   : generic
           crashedX    : process crashed
           manualX     : manual UI error
           decisionX   : automated decision errro
           workflowX   : workflow error
           badNameX   : unknown WFE process reference
           timeoutX   : process ran out of time
           startX : process never started
           all    : anything else
           rest   : anything else
        """

        logger.info("+mainEngine.__manageException:  starting")

        if self.exception is None:
            logger.info("+mainEngine.__manageException : No exception - should not be in manageException\n")
            return 0, workingTask

        if workingTask is None:
            logger.info("+mainEngine.__manageException :  Unknown current task \n")
            return -1, None

        if opt < 0:
            logger.info("+mainEngine.__manageException :  Unknown parallel number in task %d \n", opt)
            # an old way of handling exceptions - reset and continue
            opt = 0

        if workingTask.exception is None:
            logger.info("+mainEngine.__manageException  :  In task = %s", str(workingTask.name))
            logger.info("+mainEngine.__manageException :  No exception handler declared ")

        if self.debug > 0:
            logger.info("+mainEngine.__manageException  :  Exception managed = %s", str(self.exception))
            logger.info("+mainEngine.__manageException  :  In task = %s", str(workingTask.name))
            logger.info("+mainEngine.__manageException  :  Exception handler = %s", str(workingTask.exception))

        #  get the handled exception task and continue
        taskException = self.__findTaskByName(workingTask.exception, taskObjList)

        # add to the task path : need to add to the status DB
        self.path.append(taskException)

        if taskException.data is None or taskException.data == []:
            # no handlers defined
            logger.info("+mainEngine.__manageException  :  No handlers defined %s", str(self.exception))
            logger.info(str(taskException.outputName))
            if taskException.outputName is None or taskException.outputName == [] or taskException.outputName[0] == "":
                logger.info("+mainEngine.__manageException  :  No exception nexttask - will die\n")
                self.exception = "exception"
                return -1, None
            else:
                logger.info("+mainEngine.__manageException  :  Global exception next task %s \n", taskException.outputName[0])
                self.exception = None
                return 0, taskException
        else:
            # handler data is block of 5 data : print string, exception , nexttask , ,
            if self.debug > 1:
                logger.info("+mainEngine.__manageException  :  List of declared handles for this exception manager\n")
                for datum in taskException.data:
                    if datum[0] is not None:
                        logger.info(" Exception handler %s warning : %s, %s", str(taskException.name), str(datum[0]), str(datum[1]))
            # look for specific handles
            for datum in taskException.data:
                if datum[1].lower() == self.exception.lower():
                    if self.debug > 0:
                        logger.info("+mainEngine.__manageException  : Exception manager - found handler name = %s\n", str(datum[1]))
                    if datum[2].lower() == "die":
                        logger.info("+mainEngine.__manageException  :  handle to DIE called\n")
                        self.exception = "exception"
                        return -1, None
                    else:
                        taskException.outputName = []
                        taskException.outputName.append(datum[2])
                        self.exception = None
                        return 0, taskException
                # now handle any remaining options
                if datum[1].lower() == "all" or datum[1].lower() == "rest":
                    if self.debug > 0:
                        logger.info("+mainEngine.__manageException  : Exception manager - handler name = %s", str(datum[1]))
                    if datum[2].lower() == "die":
                        logger.info("+mainEngine.__manageException  :  handle to DIE called\n")
                        self.exception = "exception"
                        return -1, None
                    else:
                        taskException.outputName = []
                        taskException.outputName.append(datum[2])
                        self.exception = None
                        return 0, taskException

        return -1, None

    def runNoThrow(self, argv, runTime=None):

        self.run(argv, runTime, throwExit=False)

    def run(self, argv, runTime=None, throwExit=True):
        """
        The main workflow engine method
        1) get runtime parameters, set the main engine parameteres
        2) read the workflow class  WF
        3) Set the state for class and instance in DB
        4) Determine the task recovery point if required : fetch state
        5) run workflow until exception/end point
           Exception : run handler - and continue in WF at handle
        """

        if runTime is None:
            self.runTimeParameters = CommandLineArgs(argv)
        else:
            self.runTimeParameters = runTime

        if not self.runTimeParameters:
            return

        self.__depositionId = self.runTimeParameters.getSessionID()

        if self.runTimeParameters.debug >= 0:
            self.debug = self.runTimeParameters.debug

        self.__wfClassFileName = self.runTimeParameters.workflow
        self.__wfInstId = self.runTimeParameters.instanceID
        recoveryFlag = 0

        logger.info(
            "+mainEngine.run  ------------------------- Starting for depositionID = %s  WFInstID %s workflow file %s ",
            str(self.__depositionId),
            str(self.__wfInstId),
            str(self.__wfClassFileName),
        )

        if not self.__depositionId:
            logger.info("+mainEngine.run :  Workflow cannot proceed without a depositionID\n")
            return
        wfFilePath = os.path.join(self.runTimeParameters.getPath(), self.runTimeParameters.getWorkFlowFileName())
        parse, xmldoc = self.__readWorkFlow(wfFilePath)

        # get the meta data from the workflow class file -  sets global workflow details (e.g. classID )
        wfMetaDataObj = self.__readWorkFlowHeader(parse, xmldoc)
        self.__wfClassId = wfMetaDataObj.getID()

        # create DB entry for the class
        self.__createDBclass(wfMetaDataObj)

        # is there a parameter that will override the current workflow state
        startTask = None
        taskName = self.runTimeParameters.getInitTask()
        #
        logger.info("+mainEngine.run : command-line args class %s starting task name %s instance %s", self.__wfClassId, taskName, self.__wfInstId)

        if not taskName:
            logger.info("+mainEngine.run  : Error : cannot find task name for start point in command-line arguments\n")
            logger.info("+mainEngine.run  ------------------------------------------  returning for %s ", self.__depositionId)
            sys.exit(0)
        else:
            if taskName.lower() == "start" or taskName.lower() == "entry-point":
                # this is the normal case :
                # start from scratch with new instance ID : so create a new DB entry for this
                self.__createNewDBwfInstance()
                logger.info("+mainEngine.run  : starting workflow %s at task %s\n", self.__wfClassFileName, taskName)
            elif taskName.lower() == "recover":
                if self.__wfInstId is None:
                    logger.info("+mainEngine.run  : *** Error : Recover requires a valid WorkflowInstanceID")
                    logger.info("+mainEngine.run  ------------------------------------------  returning for %s ", self.__depositionId)
                    sys.exit(0)
                if self.__wfInstId == "recover":
                    recoveryFlag = 2
                    # find the last instanceID and the last finished task
                    #  WE NEED A TASK THAT WE CAN RECOVER FROM - get state of workflow variable
                    ws = WorkflowSession(self.__eUtil, self.__depositionId, self.__wfClassId, debug=self.debug)
                    self.__wfInstId, taskName = ws.autoRecover()
                    logger.info("+mainEngine.run %s autorecover() returns instance %s  task name %s recoveryFlag %s ", self.__depositionId, self.__wfInstId, taskName, recoveryFlag)
                elif self.__wfInstId.startswith("recoverannotate"):
                    ws = WorkflowSession(self.__eUtil, self.__depositionId, self.__wfClassId, debug=self.debug)
                    taskName = ws.recoverFromAnnotate()
                    self.__wfInstId = "Annotate"
                    recoveryFlag = 2
                    # task name is correct
                    logger.info(
                        "+mainEngine.run %s recoverFromAnnotate() returns instance %s  task name %s recoveryFlag %s ", self.__depositionId, self.__wfInstId, taskName, recoveryFlag
                    )
                else:
                    # find the last finished task
                    ws = WorkflowSession(self.__eUtil, self.__depositionId, self.__wfClassId, debug=self.debug)
                    taskName = ws.autoRecoverInstance(self.__wfInstId)
                    recoveryFlag = 1
                    logger.info(
                        "+mainEngine.run %s autoRecoverInstance() returns instance %s  task name %s recoveryFlag %s ", self.__depositionId, self.__wfInstId, taskName, recoveryFlag
                    )
            else:
                # Just get the last instance as we have a taskName
                # we passed the classID - so we have to find the actual ID
                # #
                logger.info("+mainEngine.run - searching for last instance of class %s taskName %r", self.__wfClassId, taskName)
                ws = WorkflowSession(self.__eUtil, self.__depositionId, self.__wfClassId, debug=self.debug)
                self.__wfInstId, _junk = ws.autoRecover()
                logger.info("+mainEngine.run - autorecover() class %s taskName %r returns instance %r ", self.__wfClassId, taskName, self.__wfInstId)

        #
        # - read the workflow body
        curTaskObjList, curDataObjList = self.__readWorkFlowBody(parse, xmldoc, self.__depositionId, self.__wfInstId)

        # Find the task to start the workflow from
        if taskName is None:
            logger.error("+mainEngine.run  classID %r instanceID %r  missing taskName", self.__wfClassId, self.__wfInstId)
            logger.error("+mainEngine.run  ------------------------------------------  exiting for %s ", self.__depositionId)
            sys.exit(0)
        elif taskName.lower() == "start" or taskName.lower() == "entry-point":
            startTask = self.__findEntryPoint(curTaskObjList)
        else:
            #  try the HumanReadable first
            startTask = self.__findTaskNameFromHumanName(taskName, curTaskObjList)
            if startTask is None:
                startTask = self.__findTaskByName(taskName, curTaskObjList)

        # no start task bail
        if not startTask:
            logger.info("+mainEngine.run  - returning for depositionId %s with NO startpoint in workflow %s task name %r ", self.__depositionId, self.__wfClassFileName, taskName)
            return

        logger.info(
            "+mainEngine.run - assigning start task for class %s  instance %r taskName %r startTask %r recoveryFlag %s",
            self.__wfClassId,
            self.__wfInstId,
            taskName,
            startTask,
            recoveryFlag,
        )
        self.path.append(startTask)

        workingTask = startTask
        logger.info("+mainEngine.run - Workflow starts with class %s task %s", self.__wfClassId, workingTask.name)

        if workingTask.name == "End":
            logger.info("+mainEngine.run  : depositionID %s workingtask is at END", str(self.__depositionId))
            logger.info("+mainEngine.run  ------------------------------------------  returning for %s ", self.__depositionId)
            return

        if recoveryFlag == 0:
            self.__setDBInstStatus("workflow", "running", workingTask)

        while True:
            logger.info("+mainEngine.run - begin task loop class %s instance %s workingtask %s type %s\n\n", self.__wfClassId, self.__wfInstId, workingTask.name, workingTask.type)

            opt = self.__handleTask(workingTask, curDataObjList, recoveryFlag)

            logger.info("+mainEngine.run - workingtask %s handleTask return code %r exception %r", workingTask.name, opt, self.exception)
            recoveryFlag = 0

            if self.exception is not None:
                # This will return the exception task - and where to go next if handled
                opt, workingTask = self.__manageException(workingTask, opt, curTaskObjList)

            if workingTask is None:
                break

            if workingTask.type == "Exit-point":
                break

            # get the next task from the current task
            workingTask = self.__getNextTask(workingTask, opt, curTaskObjList)

            if not workingTask:
                break

            # ie you cannot run "end" in parallel with any other task

            self.path.append(workingTask)

        if self.exception is None:
            self.__setDBInstStatus("all", self.instanceExitState, None)
        else:
            self.__eUtil.setException(self.__depositionId)
            logger.info(
                "+mainEngine.run - exception for workflow %s for %s with exception %r status %r",
                str(self.__wfClassFileName),
                self.__depositionId,
                self.exception,
                self.instanceExitState,
            )
            logger.info("+mainEngine.run  ------------------------------------------  returning for %s ", self.__depositionId)

            if throwExit:
                # set the communication table to exception
                return 1
            else:
                return

        if self.debug >= 0:
            self.__outputPath()

        logger.info("+mainEngine.run  : Completed workflow from %s for %s", str(self.__wfClassFileName), self.__depositionId)
        logger.info("+mainEngine.run  ------------------------------------------  returning for %s\n", self.__depositionId)
        if throwExit:
            return 0
        else:
            return

    def __isException(self, name):
        values = ["exception", "crashedX", "manualX", "decisionX", "workflowX", "badNameX", "timeoutX", "startX", "stopX", "loopX"]
        if name in values:
            return True
        else:
            return False


def main(argv):
    runTimeParameters = CommandLineArgs(argv)
    debugN = runTimeParameters.debug
    logPath = runTimeParameters.log

    # Set logging details --
    # logger = logging.getLogger(name='root')
    if debugN > 2:
        logger.setLevel(logging.DEBUG)
    elif debugN > 0:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.ERROR)

    logging.captureWarnings(True)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(module)s - %(message)s")
    handler = logging.FileHandler(logPath)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    ##
    #   Log is overwritten by the logging facility  so don't open here -- jdw
    #    if (log is not None):
    #        output = open(log, "w")
    #
    engine = mainEngine(debug=debugN, prt=None)

    stat = engine.run(argv, runTimeParameters)

    exit(stat)


if __name__ == "__main__":
    main(sys.argv[1:])



#
import logging
import sys
from past.builtins import basestring

logger = logging.getLogger(name="root")


class MyLogger(object):
    def __init__(self, level=logging.INFO):
        """
        Python log levels:
           CRITICAL 50
           ERROR   40
           WARNING 30
           INFO    20
           DEBUG   10
           NOTSET  0
        """
        self.__myLevel = level
        self.__currentLevel = logger.getEffectiveLevel()
        #
        #        logger.info("+MyLogger().__init__ with my level %d and current level %d" % (self.__myLevel, self.__currentLevel))
        # my level >= current level to write message

    def write(self, str):  # pylint: disable=redefined-builtin
        try:
            if self.__myLevel >= self.__currentLevel:
                # f = sys._current_frames().values()[0]
                f = sys._getframe()  # pylint: disable=protected-access
                # fN = f.f_back.f_globals['__file__']
                tL = f.f_back.f_globals["__name__"].split(".")
                mN = "      " + tL[-1]
                if (str is not None) and isinstance(str, basestring) and (str[-1] == "\n"):
                    logger.log(self.__myLevel, mN + " " + str[:-1])
                else:
                    logger.log(self.__myLevel, mN + " " + str)
        except:  # noqa: E722  pylint: disable=bare-except
            logger.exception(str)

    def flush(self):
        f = sys._current_frames().values()[0]  # pylint: disable=protected-access
        # fN = f.f_back.f_globals['__file__']
        mN = f.f_back.f_globals["__name__"]
        logger.log(self.__myLevel, mN + " ----")  # pylint: disable=logging-not-lazy



##
# File: EngineUtils.py
# Date: Mar 8, 2015
#
# Updates:
#    26-Apr-2015  jdw remove dependence on workflow metadata object
#                     remove unnecessary update of wf_class_dir table --
#     4-Aug-2015  jdw add getNextInstanceId(dataSetId)
#     9-Dec-2015  jdw add version specific path v-152 dependency formdata.pkl.  Not a future issue.
# #
import time
import sys
import os
import pickle
import datetime
import random
import smtplib

from wwpdb.utils.wf.process.ProcessRunner import ProcessRunner
from wwpdb.utils.wf.WfDataObject import WfDataObject
from wwpdb.utils.wf.dbapi.WfDbApi import WfDbApi
from wwpdb.utils.wf.dbapi.dbAPI import dbAPI

from email.mime.text import MIMEText
from wwpdb.utils.config.ConfigInfo import ConfigInfo, getSiteId
from wwpdb.apps.wf_engine.wf_engine_utils.time.TimeStamp import TimeStamp

import logging
from wwpdb.apps.wf_engine.wf_engine_utils.run.MyLogger import MyLogger

logger = logging.getLogger(name="root")

##
# JDW  -- This class will replace the DBStatusApi structure -
# #


class EngineUtils(WfDbApi):
    def __init__(self, verbose=True):
        #  This will create a database connection in the base class --
        #
        super(EngineUtils, self).__init__(log=MyLogger(level=logging.INFO), verbose=False)
        #
        self.__verbose = verbose
        self.__number = 0
        self.__timeStamp = TimeStamp()
        self.__when = self.__timeStamp.getSecondsFromReference()
        self.__lfh = MyLogger(level=logging.INFO)
        self.__SKIP_EMAIL = False
        self.__siteId = getSiteId(defaultSiteId="WWPDB_DEPLOY_PRODUCTION_RU")
        self.__cI = ConfigInfo(self.__siteId)
        if self.__siteId == "WWPDB_DEPLOY_MACOSX":
            # Always skip e-mail notifications on the development platform -
            self.__SKIP_EMAIL = True
        else:
            self.__SKIP_EMAIL = False
        logger.info("+EngineUtils.__init__ Starting ----------------------- (%s) ---------------------------------", self.__siteId)

    def updateConnection(self, timeOutSeconds=3600):
        """Update database connection if down or expired (older than timeOutSeconds)."""

        if (not self.isConnected()) or ((self.__timeStamp.getSecondsFromReference() - self.__when) > timeOutSeconds):
            ok = self.reConnect()
            if not ok:
                logger.info("+EngineUtils.checkConnection : database connection failed at connenction number = %d\n", self.__number)
                return False
            self.__when = self.__timeStamp.getSecondsFromReference()
            self.__number += 1
            logger.info("+EngineUtils.checkConnection : creating new database connection number = %d", self.__number)
        return True

    def resetInitialStateDB(self, depID):
        """
        Method to reset a deposition back to init state by setting the owner, status and class
        """

        # select ordinal from wf_instance where dep_set_id = 'D_057171'  order by status_timestamp desc limit 1;
        # get the last row in the instance table for this ID
        #
        sql = "select ordinal from wf_instance where dep_set_id = '" + str(depID) + "'  order by status_timestamp desc limit 1"
        allList = self.runSelectSQL(sql)

        if allList is not None:
            for allRow in allList:
                ordinal = allRow[0]
                logger.info("Ordinal found for ID %s , ordinal = %s", str(depID), str(ordinal))
                sql = "update wf_instance set owner = 'Annotation.bf.xml', wf_class_id = 'Annotate' , inst_status = 'init' where ordinal = " + str(ordinal)
                ok = self.runUpdateSQL(sql)
                # update the wf_instance_last table
                sql = "update wf_instance_last set owner = 'Annotation.bf.xml', wf_class_id = 'Annotate' , inst_status = 'init' where dep_set_id = '" + str(depID) + "'"
                ok = self.runUpdateSQL(sql)
                if int(ok) == 1:
                    logger.info("Updated 1 row to initialise status DB")
                else:
                    logger.info("DID NOT UPDATE status")

    def waitTime(self, sec):
        time.sleep(sec)

    def getLogDirectoryPath(self, depID):

        topSessionPath = self.__cI.get("SITE_WEB_APPS_TOP_SESSIONS_PATH")
        logDir = os.path.join(topSessionPath, "wf-logs", depID)

        if not os.path.exists(logDir):
            os.makedirs(logDir)

        return logDir

    def setException(self, depID):

        self.setCommunication("EXCEPTION", "EXCEPTION", depID)

    def setFinished(self, depID):

        self.setCommunication("FINISHED", "FINISHED", depID)

    def setCommunication(self, status, activity, depID):

        timeNow = self.__timeStamp.getSecondsFromReference()
        sql = (
            "update communication set status = '"
            + str(status)
            + "', activity = '"
            + str(activity)
            + "', actual_timestamp = "
            + str(timeNow)
            + " where dep_set_id  = '"
            + str(depID)
            + "'"
        )

        ok = self.runUpdateSQL(sql)

        if ok == 1:
            logger.info("Set the communication to %s for dep_set_id = %s", str(status), str(depID))
        else:
            logger.info("Failed to set the communication table to %s for dep_set_id = %s", str(status), str(depID))

    def testFilesExist(self, inputs, outputs):

        dinput = None
        output = None
        if inputs is not None:
            for key, value in inputs.items():
                if self.__verbose:
                    logger.info("+EngineUtils.testFilesExist :  testing input named %s content type %s", str(key), str(type(value)))
                dinput = value
        if outputs is not None:
            for key, value in outputs.items():
                if self.__verbose:
                    logger.info("+EngineUtils.testFilesExist :   setting output named %s content type %s", str(key), str(type(value)))
                output = value

        if dinput is None:
            logger.info("+EngineUtils.testFilesExist :   missing input data object")
        if output is None:
            logger.info("+EngineUtils.testFilesExist :  missing output data object ")

        fp = dinput.getFilePathReference()
        if dinput.getFilePathExists(fp):
            output.setValue("true")
        else:
            output.setValue("false")

        logger.info("+EngineUtils.testFilesExist : returns " + output.getValue() + " for " + fp)

    ##
    #  These methods are maintained for backward compatibility but should be deprecated !!!
    def setReleaseStatus(self, depID, status):

        sql = "update deposition set author_release_status_code = '" + status + "' where dep_set_id = '" + depID + "'"
        _ok = self.runUpdateSQL(sql)  # noqa: F841

    def WFEsetAnnotator(self, depID, annotator):

        sql = "update deposition set annotator_initials = '" + annotator + "' where dep_set_id = '" + depID + "'"
        _ok = self.runUpdateSQL(sql)  # noqa: F841

    def setRandomAnnotator(self, depID):

        sql = "select initials from da_users where length(user_name) = 2"
        allList = self.runSelectSQL(sql)

        n = random.randrange(1, len(allList))

        annotator = allList[n][0]

        sql = "update deposition set annotator_initials = '" + annotator + "' where dep_set_id = '" + depID + "'"
        _ok = self.runUpdateSQL(sql)  # noqa: F841

    def __sendEmail(self, email, frm, subject, message):

        if email is None:
            logger.error("+EngineUtils.__sendEmail Invalid email %s", str(email))
            return

        msg = MIMEText(message)
        msg["Subject"] = subject
        msg["From"] = frm
        msg["To"] = email
        noreplyaddr = self.__cI.get("SITE_NOREPLY_EMAIL", "noreply@mail.wwpdb.org")

        # Send the message via our own SMTP server, but don't include the
        # envelope header.
        logger.info("+EngineUtils.__sendEmail send email from %s", str(frm))
        logger.info("+EngineUtils.__sendEmail send email to %s", str(email))
        try:
            s = smtplib.SMTP("localhost")
            s.sendmail(noreplyaddr, [email], msg.as_string())
            s.quit()
        except Exception as e:
            logger.error("+EngineUtils.__sendEmail failing with %s", str(e))

    def sendDepositorEmail(self, depID, data):
        #  JDW - for compatibility WFE running under a host based site id needs to determine the e-mail path for each case -
        if os.access(os.path.join(self.__cI.get("SITE_ARCHIVE_STORAGE_PATH"), "deposit", "temp_files", "deposition-v-200", str(depID), "citation.pkl"), os.R_OK):
            url = self.__cI.get("SITE_CURRENT_DEP_EMAIL_URL")
            logger.info("+EngineUtils.sendDepositorEmail - %s  site %s return v200 url %s", depID, self.__siteId, url)
        # elif os.access(os.path.join(self.__cI.get('SITE_ARCHIVE_STORAGE_PATH'), 'deposit', 'temp_files', 'deposition-v-152', str(depID), "formdata.pkl"), os.R_OK):
        #     url = self.__cI.get('SITE_LEGACY_DEP_EMAIL_URL')
        #     logger.info("+EngineUtils.sendDepositorEmail - %s  site %s return v152 url %s", depID, self.__siteId, url)
        elif os.access(os.path.join(self.__cI.get("SITE_ARCHIVE_STORAGE_PATH"), "deposit", "temp_files", "deposition", str(depID), "formdata.pkl"), os.R_OK):
            url = self.__cI.get("SITE_DEP_EMAIL_URL")
            logger.info("+EngineUtils.sendDepositorEmail - %s  site %s return legacy url %s", depID, self.__siteId, url)
        else:
            url = self.__cI.get("SITE_DEP_EMAIL_URL")
            logger.info("+EngineUtils.sendDepositorEmail path tests failing - %s  site %s using default url %s", depID, self.__siteId, url)
        #

        data = data.replace("$DEPID", depID)
        data = data.replace("$DEPURL", url)
        data = data.replace("$LINEFEED", "\n")

        words = data.split("|")
        email = self.__getDepositorEmail(depID)
        logger.info("+EngineUtils.sendDepositorEmail - %s sending to email %s ", depID, email)
        if self.__SKIP_EMAIL:
            logger.info("+EngineUtils.sendDepositorEmail - Skipped sending email to %s ", email)
        else:
            self.__sendEmail(email, words[0], words[1], words[2])

    def __getDepositorEmail(self, depID):
        """
        go get the depositor email

        ## JDW modified to work in V152/V200 compatibility mode -
        """
        fName = os.path.join(self.__cI.get("SITE_ARCHIVE_STORAGE_PATH"), "deposit", "temp_files", "deposition", str(depID), "formdata.pkl")
        if not os.access(fName, os.R_OK):
            fName = os.path.join(self.__cI.get("SITE_ARCHIVE_STORAGE_PATH"), "deposit", "temp_files", "deposition-v-152", str(depID), "formdata.pkl")
        #
        logger.info("+EngineUtils.__getDepositorEmail - %s at %s searching for email in file %s ", depID, self.__siteId, fName)
        if os.path.isfile(fName):
            try:
                logger.info("+EngineUtils.__getDepositorEmail - %s s opening serialized file %s ", depID, fName)
                f = open(fName, "rb")
                dat = pickle.load(f)
                f.close()
                if "email" in dat:
                    return dat["email"]
                else:
                    return None
            except Exception as e:
                logger.exception("+EngineUtils.__getDepositorEmail failed to recover formdata and return email  %s", str(e))
                return None
        else:
            logger.info("+EngineUtils.__getDepositorEmail - %s searching for email in database", depID)
            ss = dbAPI(depID)
            ret = ss.runSelectNQ(table="user_data", select=["email", "role"], where={"dep_set_id": depID, "role": "valid"})
            for r in ret:
                return r[0]
            return None

    def initDepositContext(self, depositionID, WorkflowClassID, WorkflowInstanceID, dinput):
        """
        method to extract information from the current model file and load this into the status database.
        """

        logger.info("+EngineUtils.initDepositContext : ============================================")
        logger.info("+EngineUtils.initDepositContext : depID = %s", str(depositionID))
        logger.info("+EngineUtils.initDepositContext : ============================================")

        if self.__verbose:
            process = ProcessRunner(True, self.__lfh)
        else:
            process = ProcessRunner(False, self.__lfh)

        process.setAction("status")

        if dinput is None:
            logger.info("+EngineUtils.initDepositContext :  Requires input object \n")
            sys.exit(0)

        for key, value in dinput.items():
            if self.__verbose:
                logger.info("+EngineUtils.initDepositContext :  setting input %s %s", str(key), str(value.getReferenceType()))
                logger.info("+EngineUtils.initDepositContext :  setting input %s %s", str(key), str(value))
            process.setInput("src", value)

        wfoOut = WfDataObject()
        wfoOut.setContainerTypeName("dict")
        wfoOut.setValueTypeName("string")
        process.setOutput("dst", wfoOut)

        if not process.preCheck():
            logger.info("+EngineUtils.initDepositContext  : failed to setup process\n")

        if not process.run():
            logger.info("+EngineUtils.initDepositContext : failed to run process\n")

        dataDict = wfoOut.getValue()
        if dataDict is None:
            logger.info(" There is no dataDictionary - exiting")
            exit(0)

        depDB = {}
        depDB["DEP_SET_ID"] = depositionID
        depDB["PDB_ID"] = "????"
        if "accessions" in dataDict:
            acc = dataDict["accessions"]
            depDB["PDB_ID"] = "????"
            if acc is not None:
                accPDB = acc["PDB"]
                if accPDB is not None:
                    if len(accPDB) > 0:
                        depDB["PDB_ID"] = accPDB[0]

        if "recvd_initial_deposition_date" in dataDict:
            depDB["INITIAL_DEPOSITION_DATE"] = dataDict["recvd_initial_deposition_date"]
        else:
            # JDW JDW
            today = datetime.datetime.utcnow().strftime("%Y-%m-%d")
            # depDB['INITIAL_DEPOSITION_DATE'] = 'unknown'
            depDB["INITIAL_DEPOSITION_DATE"] = today
        if "deposit_site" in dataDict:
            depDB["DEPOSIT_SITE"] = dataDict["deposit_site"]
        else:
            depDB["DEPOSIT_SITE"] = "unknown"
        if "process_site" in dataDict:
            depDB["PROCESS_SITE"] = dataDict["process_site"]
        else:
            depDB["PROCESS_SITE"] = "unknown"
        if "status_code" in dataDict:
            depDB["STATUS_CODE"] = dataDict["status_code"]
        else:
            depDB["STATUS_CODE"] = "PROC"

        logger.info("******* OVERWRITE OF STATUS CODE to PROC ***************")
        depDB["STATUS_CODE"] = "PROC"
        if "annotator_initials" not in dataDict or dataDict["annotator_initials"] == "UNASSIGNED":
            depDB["ANNOTATOR_INITIALS"] = "unknown"
        else:
            depDB["ANNOTATOR_INITIALS"] = dataDict["annotator_initials"]
            logger.info("******* OVERWRITE OF ANNOTATOR INITIALS TO AN *******")
            depDB["ANNOTATOR_INITIALS"] = "unknown"
        if "author_release_status_code" not in dataDict:
            depDB["AUTHOR_RELEASE_STATUS_CODE"] = "HPUB"
        else:
            depDB["AUTHOR_RELEASE_STATUS_CODE"] = dataDict["author_release_status_code"]
        depDB["TITLE"] = dataDict["title"].replace("'", " ")

        if "audit_author" in dataDict:
            auth = dataDict["audit_author"]
        else:
            auth = "unknown"

        if auth is not None:
            if isinstance(auth, list):
                depDB["AUTHOR_LIST"] = ""
                depDB["AUTHOR_LIST"] = " ".join(["%s," % a for a in auth])
            elif isinstance(auth, str):
                depDB["AUTHOR_LIST"] = auth
            else:
                logger.info(" Unknown type of author list %s", str(type(auth)))
            depDB["AUTHOR_LIST"] = depDB["AUTHOR_LIST"].replace("'", " ")
        else:
            depDB["AUTHOR_LIST"] = "unknown"
        #
        depDB["EXP_METHOD"] = dataDict["exp_method"]
        if depDB["EXP_METHOD"].find("NMR") >= 0:
            if "status_code_nmr" in dataDict:
                depDB["STATUS_CODE_EXP"] = dataDict["status_code_nmr"]
            else:
                depDB["STATUS_CODE_EXP"] = "Unknown"
        elif depDB["EXP_METHOD"].find("EM") >= 0:
            if "status_code_em" in dataDict:
                depDB["STATUS_CODE_EXP"] = dataDict["status_code_em"]
            else:
                depDB["STATUS_CODE_EXP"] = "Unknown"
        else:
            if "status_code_sf" in dataDict:
                depDB["STATUS_CODE_EXP"] = dataDict["status_code_sf"]
            else:
                depDB["STATUS_CODE_EXP"] = "Unknown"
        if "full_name_of_center" in dataDict:
            depDB["SG_CENTER"] = dataDict["full_name_of_center"]
        else:
            depDB["SG_CENTER"] = "Unknown"

        if self.exist(depDB):
            # then we do nothing - since this workflow does not modify the depostion
            constDict = {}
            constDict["DEP_SET_ID"] = depositionID
            self.saveObject(depDB, "update", constDict)
        else:
            self.saveObject(depDB, "insert")
            if self.__verbose:
                logger.info("+EngineUtils.initDepositContext : done insert data object in DB\n")

        instDB = {}
        instDB["WF_INST_ID"] = WorkflowInstanceID
        instDB["WF_CLASS_ID"] = WorkflowClassID
        instDB["DEP_SET_ID"] = depositionID
        instDB["STATUS_TIMESTAMP"] = datetime.datetime.utcnow()
        self.updateStatus(instDB, "init")

        # and also initialise the wf_instance_last table
        sql = "update wf_instance_last set status_timestamp=" + str(self.__timeStamp.getSecondsFromReference()) + ", inst_status='init' where dep_set_id = '" + depositionID + "'"
        ok = self.runUpdateSQL(sql)
        if ok < 1:
            logger.error("+EngineUtils.initDepositContext() failed to update wf_instance_last table")

    def getNextInstanceId(self, dataSetId):
        """
        Create new/next workflow instance for the input data set Id using wf_instance table data.

        """

        sql = "select max(cast(substr(wf_inst_id,3) as decimal(5,0)))+1 from wf_instance where dep_set_id = '" + str(dataSetId) + "'"

        ll = self.runSelectSQL(sql)

        if ll is None:
            instID = 1
        else:
            for l in ll:  # noqa: E741
                if l is None:
                    instID = 1
                else:
                    if l[0] is None:
                        instID = 1
                    else:
                        instID = l[0]

        ret = "W_" + str(instID).zfill(3)

        if self.__verbose > 0:
            logger.info("+mainEngine.__getNewInstanceIDDB : for data set ID = %s next workflow instance ID = %s", dataSetId, ret)

        return ret


##
#  END JDW



from wwpdb.utils.wf.dbapi.DbApiUtil import DbApiUtil
from wwpdb.utils.config.ConfigInfo import ConfigInfo
import sys


class AutoRun(object):
    __queries = {
        "LAST_STATUS_AND_ORDINAL": ("select status, ordinal from communication " "where parent_dep_set_id = '{dep_id}' " "order by actual_timestamp desc limit 1"),
        "START_ANNOTATION_WORKFLOW": (
            "UPDATE communication set sender='WFM', receiver='WFE', dep_set_id='{dep_id}', "
            "wf_class_file='Annotation.bf.xml', command='runWF', status='pending', "
            "parent_dep_set_id='{dep_id}', parent_wf_class_id='Annotate' "
            "WHERE ordinal = {ordinal}"
        ),
    }

    def __init__(self, dep_id, site_id=None, verbose=False, log=sys.stderr):
        """ """
        self.__lfh = log
        self.__verbose = verbose
        self.__siteId = site_id
        self.dep_id = dep_id
        self.db_api = self.make_db_connection()
        self.wf_status = None
        self.ordinal = None

    def make_db_connection(self):
        config_info = ConfigInfo(self.__siteId)

        db_server = config_info.get("SITE_DB_SERVER")
        db_host = config_info.get("SITE_DB_HOST_NAME")
        db_name = config_info.get("SITE_DB_DATABASE_NAME")
        db_user = config_info.get("SITE_DB_USER_NAME")
        db_pw = config_info.get("SITE_DB_PASSWORD")
        db_socket = config_info.get("SITE_DB_SOCKET")
        db_port = int(config_info.get("SITE_DB_PORT_NUMBER"))

        return DbApiUtil(dbServer=db_server, dbHost=db_host, dbName=db_name, dbUser=db_user, dbPw=db_pw, dbSocket=db_socket, dbPort=db_port, verbose=self.__verbose, log=self.__lfh)

    def get_status_and_ordinal(self):
        sql = self.__queries["LAST_STATUS_AND_ORDINAL"].format(dep_id=self.dep_id)
        rows = self.db_api.runSelectSQL(sql)
        for row in rows:
            self.wf_status = row.get("status", None)
            self.ordinal = row.get("ordinal", None)
            break
        else:
            self.__lfh.write("Couldn't get Status of current workflow")

    def run_annotation(self):
        sql = self.__queries["START_ANNOTATION_WORKFLOW"].format(dep_id=self.dep_id, ordinal=self.ordinal)
        ret = self.db_api.runUpdateSQL(sql)
        if ret != "OK":
            self.__lfh.write("Start Annotate workflow failed.")
        else:
            self.__lfh.write("Annotation Workflow Started.")
